from abc import abstractmethod
import os


class Function(object):
    """Interface for Pulsar Function"""
    @abstractmethod
    def process(self, input, context):
        """Process input message"""
        pass


class Context:
    _data = {
        '_user': {}
    }

    def set_user_config_value(self, key, value):
        self._data['_user'][key] = value

    def get_user_config_value(self, key):
        if key in self._data['_user']:
            return self._data['_user'][key]
        else:
            return None

    def get_user_config_map(self):
        return self._data['_user']

    def get_logger(self):
        return Logger()

    def put_state(self, key, value):
        if not isinstance(value, str):
            raise Exception(f"Value passed in to store is not a string.")
        f = open(f"/tmp/{key}", "w")
        f.write(value)
        f.close()

    def get_state(self, key):
        if os.path.exists(f"/tmp/{key}"):
            f = open(f"/tmp/{key}", "r")
            data = f.read()
            f.close()
            return data

        return None

    def get_counter(self, key):
        try:
            return self._data[key]
        except:
            return None

    def del_counter(self, key):
        return self._data.pop(key, None)

    def incr_counter(self, key, amount):
        if key not in self._data:
            self._data[key] = amount
        else:
            self._data[key] += amount

    def publish(self, topic_name, message, serde_class_name="serde.IdentitySerDe",
                              properties=None, compression_type=None, callback=None, message_conf=None):
        print(f"Simulating putting {message} on topic {topic_name}")
        if callable(callback):
            callback()


class Logger:
    def info(self, message):
        print(message)

    def error(self, message):
        print(message)

    def warning(self, message):
        print(message)

    def debug(self, message):
        print(message)


# class ContextImpl(pulsar.Context):
#   def get_message_id(self):
#     ...
#   def get_message_key(self):
#     ...
#   def get_message_eventtime(self):
#     ...
#   def get_message_properties(self):
#     ...
#   def get_current_message_topic_name(self):
#     ...
#   def get_partition_key(self):
#     ...
#   def get_function_name(self):
#     ...
#   def get_function_tenant(self):
#     ...
#   def get_function_namespace(self):
#     ...
#   def get_function_id(self):
#     ...
#   def get_instance_id(self):
#     ...
#   def get_function_version(self):
#     ...
#   def get_logger(self):
#     ...
#   def get_user_config_value(self, key):
#     ...
#   def get_user_config_map(self):
#     ...
#   def record_metric(self, metric_name, metric_value):
#     ...
#   def get_input_topics(self):
#     ...
#   def get_output_topic(self):
#     ...
#   def get_output_serde_class_name(self):
#     ...
#   def publish(self, topic_name, message, serde_class_name="serde.IdentitySerDe",
#               properties=None, compression_type=None, callback=None, message_conf=None):
#     ...
#   def ack(self, msgid, topic):
#     ...
#   def get_and_reset_metrics(self):
#     ...
#   def reset_metrics(self):
#     ...
#   def get_metrics(self):
#     ...
#   def put_state(self, key, value):
#     ...
#   def get_state(self, key):
#     ...


from abc import abstractmethod
import pickle
import time


class FileStorage:
    def save(self, directory_wo_slash=None, checkpoint=False):
        if self.get_pipeline_id() is None:
            print("Cannot save artifacts, no pipeline id found.")
            return False

        import os.path

        if directory_wo_slash is None:
            storage_location = self._storage_location + '/' + self.get_pipeline_id() + '/' + self.get_name_id()
        else:
            storage_location = directory_wo_slash + '/' + self.get_pipeline_id() + '/' + self.get_name_id()

        if checkpoint:
            storage_location += "/checkpoint"
        else:
            storage_location += "/completed"

        storage_location = storage_location + "/" + time.strftime("%Y%m%d-%H%M%S")

        print(f"Saving artifacts to {storage_location}")

        os.makedirs(storage_location)

        for item in self._artifacts.items():
            if callable(item[1]):
                print(f"Calling method for {item[0]}")
                try:
                    item[1](storage_location)
                except Exception as e:
                    print(f"An error occurred saving artifact {item[0]}: {e}")
            else:
                print(f"Saving artifact with name: {item[0]}")
                try:
                    # Don't overwrite existing file, this shouldn't happen, but make sure it does not.
                    if not os.path.isfile(f"{storage_location}/{item[0]}.pickle"):
                        with open(f"{storage_location}/{item[0]}.pickle", 'wb') as handle:
                            pickle.dump(item[1], handle)
                except Exception as e:
                    print(f"An error occurred saving artifact {item[0]}: {e}")

        return True


class S3Storage:
    def __init__(self):
        self.client = None

    def save(self):
        print(f"Using client to save ({len(self._artifacts)}) element(s).")
        return True


class Artifact(FileStorage):
    _artifacts = {}
    _pipeline_id = None
    _name_id = 'example'
    _config = None
    _storage_location = '/tmp'

    def __init__(self):
        super().__init__()

    def add(self, _unique_name, _artifact):
        self._artifacts[_unique_name] = _artifact
        return _artifact

    def set_storage_location(self, storage_location):
        self._storage_location = storage_location

    def set_pipeline_id(self, pipeline_id):
        self._pipeline_id = pipeline_id

    def set_name_id(self, name_id):
        self._name_id = name_id

    def get_pipeline_id(self):
        return self._pipeline_id

    def get_name_id(self):
        return self._name_id


artifact = Artifact()



from pandioml.data.stream import Stream
import pulsar
from pandioml.data.record import Record, String, Integer, Float, JsonSchema


class CreditCardFraud(Stream):
    """
    A dataset of 1,296,675 credit card transactions with a percentage labeled as fraud.

    Each record is an instance of Transaction

    class Transaction(Record):
        trans_date_trans_time = String()
        cc_num = Integer()
        merchant = String()
        category = String()
        amt = Float()
        first = String()
        last = String()
        gender = String()
        street = String()
        city = String()
        state = String()
        zip = Integer()
        lat = Float()
        long = Float()
        city_pop = Integer()
        job = String()
        dob = String()
        trans_num = String()
        unix_time = Integer()
        merch_lat = Float()
        merch_long = Float()
        is_fraud = Integer()
        weekday = Integer()
        weekend = Integer()
        month = Integer()
        day = Integer()
        hour = Integer()
    """
    index = 0
    dataset = None

    def __init__(self, start_id=pulsar.MessageId.earliest, receiver_queue_size = 1000, pandio_token=None):
        self.client = pulsar.Client('pulsar+ssl://joshuaeric--gray-guan.us-west2.gcp.pulsar.pandio.com:6651',
                                    authentication=pulsar.AuthenticationToken(pandio_token))

        self.reader = self.client.create_reader('persistent://public/default/credit-card-fraud', start_id,
                                                schema=JsonSchema(Transaction),
                                                receiver_queue_size=receiver_queue_size)

    def next(self):
        msg = self.reader.read_next()
        self.index += 1
        return msg

    @staticmethod
    def schema():
        return JsonSchema(Transaction)


class Transaction(Record):
    trans_date_trans_time = String()
    cc_num = Integer()
    merchant = String()
    category = String()
    amt = Float()
    first = String()
    last = String()
    gender = String()
    street = String()
    city = String()
    state = String()
    zip = Integer()
    lat = Float()
    long = Float()
    city_pop = Integer()
    job = String()
    dob = String()
    trans_num = String()
    unix_time = Integer()
    merch_lat = Float()
    merch_long = Float()
    is_fraud = Integer()
    weekday = Integer()
    weekend = Integer()
    month = Integer()
    day = Integer()
    hour = Integer()



from river.tree import splitter
from river.tree.extremely_fast_decision_tree import ExtremelyFastDecisionTreeClassifier
from river.tree.hoeffding_adaptive_tree_classifier import HoeffdingAdaptiveTreeClassifier
from river.tree.hoeffding_adaptive_tree_regressor import HoeffdingAdaptiveTreeRegressor
from river.tree.hoeffding_tree_classifier import HoeffdingTreeClassifier
from river.tree.hoeffding_tree_regressor import HoeffdingTreeRegressor
from river.tree.isoup_tree_regressor import iSOUPTreeRegressor
from river.tree.label_combination_hoeffding_tree import LabelCombinationHoeffdingTreeClassifier

__all__ = [
    "splitter",
    "HoeffdingTreeClassifier",
    "ExtremelyFastDecisionTreeClassifier",
    "HoeffdingAdaptiveTreeClassifier",
    "HoeffdingTreeRegressor",
    "HoeffdingAdaptiveTreeRegressor",
    "iSOUPTreeRegressor",
    "LabelCombinationHoeffdingTreeClassifier"
]


from river.multiclass.occ import OutputCodeClassifier
from river.multiclass.ovo import OneVsOneClassifier
from river.multiclass.ovr import OneVsRestClassifier

__all__ = ["OutputCodeClassifier", "OneVsOneClassifier", "OneVsRestClassifier"]


import pickle
import codecs
from river.utils import inspect, math, pretty, skmultiflow_utils
from river.utils.data_conversion import dict2numpy, numpy2dict
from river.utils.estimator_checks import check_estimator
from river.utils.histogram import Histogram
from river.utils.param_grid import expand_param_grid
from river.utils.sdft import SDFT
from river.utils.skyline import Skyline
from river.utils.window import SortedWindow, Window
import os

__all__ = [
    "check_estimator",
    "dict2numpy",
    "expand_param_grid",
    "inspect",
    "math",
    "pretty",
    "Histogram",
    "numpy2dict",
    "SDFT",
    "skmultiflow_utils",
    "Skyline",
    "SortedWindow",
    "Window",
    "ModelUtility"
]


class ModelUtility:
    @staticmethod
    def load_or_instantiate(path, model):
        if os.path.isfile(path):
            return pickle.loads(path)

        if callable(model):
            return model()

        return None

    @staticmethod
    def upload(model_name, model, storage):
        if model_name is not None:
            return storage.set(model_name, codecs.encode(pickle.dumps(model), "base64").decode())
        else:
            return False

    @staticmethod
    def download(model_name, storage):
        if model_name is not None:
            state = storage.get(model_name)
            if state is not None:
                return pickle.loads(codecs.decode(state.encode(), "base64"))

        return None

    @staticmethod
    def combine(model_a, model_b):
        #model_a.estimators_ += model_b.estimators_
        #model_a.n_estimators = len(model_a.estimators_)
        #return model_a
        return model_a



from river.feature_extraction.agg import Agg, TargetAgg
from river.feature_extraction.kernel_approx import RBFSampler
from river.feature_extraction.poly import PolynomialExtender
from river.feature_extraction.vectorize import TFIDF, BagOfWords

__all__ = [
    "Agg",
    "BagOfWords",
    "PolynomialExtender",
    "RBFSampler",
    "TargetAgg",
    "TFIDF",
]


from river.imblearn.hard_sampling import HardSamplingClassifier, HardSamplingRegressor
from river.imblearn.random import RandomOverSampler, RandomSampler, RandomUnderSampler

__all__ = [
    "HardSamplingClassifier",
    "HardSamplingRegressor",
    "RandomOverSampler",
    "RandomUnderSampler",
    "RandomSampler",
]


from .form_submissions import FormSubmissionGenerator, Submission
from .hosting import WebHostingDataset, ResourceEvent
from .people import PersonProfileGenerator, PersonProfile
from .agrawal import AgrawalGeneratorDataset, LoanApplication
from .sine_generator import SineGeneratorDataset, Sine
from .led_generator import LEDGeneratorDataset, LED
from .credit_card_transactions import CreditCardFraud, Transaction
from .phishing_dataset import PhisingDataset, Webpage
from .movie_ratings import MovieRatingDataset, MovieRating
from .restaurant_visitors import RestaurantVisitorsDataset, RestaurantDay
#from .scikit_multiflow_classes import BaseEstimator, BaseSKMObject, SKStream, check_random_state
#from .stream import Stream
#from .record import Record, Field, Null, Boolean, Integer, Long, Float, Double, Bytes, String, Array, Map, \
#    JsonSchema, Schema, BytesSchema, StringSchema, JsonSchema, AvroSchema

#__all__ = ["FormSubmissionGenerator", "WebHostingDataset", "PersonProfileGenerator", "Submission", "ResourceEvent",
#           "PersonProfile", "AgrawalGenerator", "CreditCardFraud", "PhisingDataset", "LoanApplication", "Transaction",
#           "Webpage", "MovieRating", "MovieRatingDataset", "Stream", "Record", "Field", "Null", "Boolean", "Integer",
#           "Long", "Float", "Double", "Bytes", "String", "Array", "Map", "JsonSchema", "Schema", "BytesSchema",
#           "StringSchema", "AvroSchema"]

__all__ = ["FormSubmissionGenerator", "WebHostingDataset", "PersonProfileGenerator", "Submission", "ResourceEvent",
           "PersonProfile", "AgrawalGeneratorDataset", "CreditCardFraud", "PhisingDataset", "LoanApplication", "Transaction",
           "Webpage", "MovieRating", "MovieRatingDataset", "RestaurantVisitorsDataset", "RestaurantDay", "Sine",
           "SineGeneratorDataset", "LEDGeneratorDataset", "LED"]



from abc import abstractmethod, ABCMeta, abstractproperty
import signal
import sys
from pandioml.core.artifacts import artifact


class FunctionBase(object, metaclass=ABCMeta):
    """Interface for Pandio Function"""
    model = abstractproperty()
    startup_ran = False
    input = None
    storage = None
    config = None
    _result = None
    save_model_after_this_many_events = 10000

    @classmethod
    def __init__(cls, input=None, context=None, config=None):
        cls.input = input
        cls.storage = Storage(context=context)
        cls.config = config

    @abstractmethod
    def pipelines(self):
        raise NotImplementedError

    @classmethod
    def get_result(self):
        return self._result

    @classmethod
    def set_result(self, result):
        self._result = result

    @classmethod
    def shutdown(cls):
        artifact.save()
        print("SHUTDOWN")

    @classmethod
    def startup(cls):
        if cls.startup_ran is False:
            print("STARTUP")

            # TODO, Pulsar runs this in a child process, so these do not work
            # Only one signal can be registered, only register if this is not running inside of runner.py
            if 1 != 1 and sys.argv[0][-9:] != 'runner.py':
                signal.signal(signal.SIGINT, cls.shutdown)
                signal.signal(signal.SIGTERM, cls.shutdown)

            cls.startup_ran = True

    @classmethod
    def fit(cls, result={}):
        cls.model.learn_one(result['features'], result['labels'])
        return result

    @classmethod
    def predict(cls, result={}):
        result['prediction'] = cls.model.predict_one(result['features'])
        # Sometimes this returns a numpy.float64
        # TODO, dig into why
        if hasattr(result['prediction'], 'item'):
            result['prediction'] = result['prediction'].item()

        return result

    @classmethod
    def error(cls, result={}):
        print(result[0])
        print(result[1])
        #raise Exception(f"An exception occurred in the pipeline: {result[0]} {result[1]}")
        raise Exception("Halt!")


class Storage:
    context = None

    def __init__(self, context=None):
        self.context = context

    def set(self, key, value):
        return self.context.put_state(key, value)

    def get(self, key):
        return self.context.get_state(key)

    def increment_counter(self, key, amount):
        return None
        return self.context.increment_counter(key, amount)

    def delete_counter(self, key):
        return None
        return self.context.del_counter(key)

    def get_counter(self, key):
        return None
        return self.context.get_counter(key)



from river.multioutput.chain import ClassifierChain, MonteCarloClassifierChain, ProbabilisticClassifierChain, \
    RegressorChain

__all__ = [
    "ClassifierChain",
    "MonteCarloClassifierChain",
    "ProbabilisticClassifierChain",
    "RegressorChain",
]


from .naive_bayes import GaussianNB, MultinomialNB, ComplementNB, BernoulliNB
from .anomaly_detection import HalfSpaceTrees
from .cluster import KMeans
from .ensemble import AdaptiveRandomForestClassifier, AdaptiveRandomForestRegressor, AdaBoostClassifier, \
    ADWINBaggingClassifier, BaggingClassifier, BaggingRegressor, LeveragingBaggingClassifier, SRPClassifier
from .evaluation import load_binary_clf_tracks, progressive_val_score, Track
from .expert import EpsilonGreedyRegressor, EWARegressor, SuccessiveHalvingClassifier, SuccessiveHalvingRegressor, \
    StackingClassifier, UCBRegressor
from .factor import FFMClassifier, FFMRegressor, FMClassifier, FMRegressor, FwFMClassifier, FwFMRegressor, \
    HOFMClassifier, HOFMRegressor
from .feature_extraction import Agg, TargetAgg, RBFSampler, PolynomialExtender, TFIDF, BagOfWords
from .feature_selection import SelectKBest, PoissonInclusion, VarianceThreshold
from .imblearn import HardSamplingClassifier, HardSamplingRegressor, RandomOverSampler, RandomSampler, RandomUnderSampler
from .linear import ALMAClassifier, LinearRegression, LogisticRegression, Perceptron, PAClassifier, PARegressor, \
    SoftmaxRegression
from .meta import PredClipper, BoxCoxRegressor, TransformedTargetRegressor
from .multiclass import OutputCodeClassifier, OneVsOneClassifier, OneVsRestClassifier
from .multioutput import ClassifierChain, MonteCarloClassifierChain, ProbabilisticClassifierChain, \
    RegressorChain
from .neighbors import KNNADWINClassifier, KNNClassifier, KNNRegressor, SAMKNNClassifier
from .neural_network import activations, MLPRegressor
from .optim import losses, initializers, schedulers ,AdaBound, AdaDelta, AdaGrad, AdaMax, Adam, AMSGrad, Averager, \
    Optimizer, FTRLProximal, Momentum, Nadam, NesterovMomentum, RMSProp, SGD
from .preprocessing import FeatureHasher, PreviousImputer, StatImputer, LDA, OneHotEncoder, \
    AdaptiveStandardScaler, Binarizer, MaxAbsScaler, MinMaxScaler, Normalizer, RobustScaler, StandardScaler
from .proba import Gaussian, Multinomial
from .reco import Baseline, BiasedMF, FunkMF, RandomNormal
from .stream import Cache, iter_arff, iter_array, iter_csv, iter_libsvm, simulate_qa, shuffle
from .time_series import Detrender, GroupDetrender, SNARIMAX
from .trees import splitter, ExtremelyFastDecisionTreeClassifier, HoeffdingAdaptiveTreeClassifier, \
    HoeffdingAdaptiveTreeRegressor, HoeffdingTreeClassifier, HoeffdingTreeRegressor, iSOUPTreeRegressor, \
    LabelCombinationHoeffdingTreeClassifier
from .utility import inspect, math, pretty, skmultiflow_utils, dict2numpy, numpy2dict, check_estimator, Histogram, \
    expand_param_grid, SDFT, Skyline, SortedWindow, Window, ModelUtility


__all__ = ["GaussianNB", "MultinomialNB", "ComplementNB", "BernoulliNB", "ModelUtility", "KNNClassifier",
           "KNNADWINClassifier", "SAMKNNClassifier", "KNNRegressor", "HoeffdingTreeClassifier",
           "HoeffdingAdaptiveTreeClassifier", "ExtremelyFastDecisionTreeClassifier",
           "LabelCombinationHoeffdingTreeClassifier", "HoeffdingTreeRegressor", "HoeffdingAdaptiveTreeRegressor",
           "iSOUPTreeRegressor", "AdaptiveRandomForestClassifier", "AdaptiveRandomForestRegressor", "ClassifierChain",
           "ProbabilisticClassifierChain", "MonteCarloClassifierChain", "LeveragingBaggingClassifier", "RegressorChain",
           "HalfSpaceTrees", "KMeans", "AdaptiveRandomForestClassifier", "AdaptiveRandomForestRegressor",
           "AdaBoostClassifier", "ADWINBaggingClassifier", "BaggingClassifier", "BaggingRegressor",
           "LeveragingBaggingClassifier", "SRPClassifier", "load_binary_clf_tracks", "progressive_val_score", "Track",
           "EpsilonGreedyRegressor", "EWARegressor", "SuccessiveHalvingClassifier", "SuccessiveHalvingRegressor",
           "UCBRegressor", "FFMClassifier", "FFMRegressor", "FMClassifier", "FMRegressor", "FwFMClassifier",
           "FwFMRegressor", "HOFMClassifier", "HOFMRegressor", "Agg", "BagOfWords", "PolynomialExtender", "RBFSampler",
           "TargetAgg", "TFIDF", "PoissonInclusion", "SelectKBest", "VarianceThreshold", "HardSamplingClassifier",
           "HardSamplingRegressor", "RandomOverSampler", "RandomUnderSampler", "RandomSampler", "ALMAClassifier",
           "LinearRegression", "LogisticRegression", "PAClassifier", "PARegressor", "Perceptron", "SoftmaxRegression",
           "BoxCoxRegressor", "PredClipper", "TransformedTargetRegressor", "OutputCodeClassifier", "OneVsOneClassifier",
           "OneVsRestClassifier", "ClassifierChain", "MonteCarloClassifierChain", "ProbabilisticClassifierChain",
           "RegressorChain", "KNNADWINClassifier", "KNNClassifier", "KNNRegressor", "SAMKNNClassifier", "activations",
           "MLPRegressor", "AdaBound", "AdaDelta" ,"AdaGrad", "Adam", "AMSGrad", "AdaMax", "Averager", "FTRLProximal",
           "initializers", "losses", "Momentum", "Nadam", "NesterovMomentum", "Optimizer", "RMSProp", "schedulers",
           "SGD", "AdaptiveStandardScaler", "Binarizer", "FeatureHasher", "LDA", "MaxAbsScaler", "MinMaxScaler",
           "Normalizer", "OneHotEncoder", "PreviousImputer", "RobustScaler", "StandardScaler", "StatImputer",
           "Gaussian", "Multinomial", "Baseline", "BiasedMF", "FunkMF", "RandomNormal", "Cache", "iter_arff",
           "iter_array", "iter_csv", "iter_libsvm", "simulate_qa", "StackingClassifier",  "SortedWindow", "Window",
           "shuffle", "Detrender", "GroupDetrender", "SNARIMAX", "splitter", "check_estimator", "dict2numpy",
           "expand_param_grid", "inspect", "math", "pretty", "Histogram", "numpy2dict", "SDFT", "skmultiflow_utils",
           "Skyline"]



from river.meta.pred_clipper import PredClipper
from river.meta.target_transform import BoxCoxRegressor, TransformedTargetRegressor

__all__ = ["BoxCoxRegressor", "PredClipper", "TransformedTargetRegressor"]


from river.preprocessing.feature_hasher import FeatureHasher
from river.preprocessing.impute import PreviousImputer, StatImputer
from river.preprocessing.lda import LDA
from river.preprocessing.one_hot import OneHotEncoder
from river.preprocessing.scale import AdaptiveStandardScaler, Binarizer, MaxAbsScaler, MinMaxScaler, Normalizer, \
    RobustScaler, StandardScaler


__all__ = [
    "AdaptiveStandardScaler",
    "Binarizer",
    "FeatureHasher",
    "LDA",
    "MaxAbsScaler",
    "MinMaxScaler",
    "Normalizer",
    "OneHotEncoder",
    "PreviousImputer",
    "RobustScaler",
    "StandardScaler",
    "StatImputer",
]


from river.optim import losses
from river.optim import initializers, schedulers
from river.optim.ada_bound import AdaBound
from river.optim.ada_delta import AdaDelta
from river.optim.ada_grad import AdaGrad
from river.optim.ada_max import AdaMax
from river.optim.adam import Adam
from river.optim.ams_grad import AMSGrad
from river.optim.average import Averager
from river.optim.base import Optimizer
from river.optim.ftrl import FTRLProximal
from river.optim.momentum import Momentum
from river.optim.nadam import Nadam
from river.optim.nesterov import NesterovMomentum
from river.optim.rms_prop import RMSProp
from river.optim.sgd import SGD

__all__ = [
    "AdaBound",
    "AdaDelta",
    "AdaGrad",
    "Adam",
    "AMSGrad",
    "AdaMax",
    "Averager",
    "FTRLProximal",
    "initializers",
    "losses",
    "Momentum",
    "Nadam",
    "NesterovMomentum",
    "Optimizer",
    "RMSProp",
    "schedulers",
    "SGD",
]


from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Integer, Float
from river.datasets import Phishing


class PhisingDataset(Stream):
    """
    This dataset contains features from web pages that are classified as phishing or not.

    Each record is an instance of Webpage

    class Webpage(Record):
        empty_server_form_handler = Float()
        popup_window = Float()
        https = Float()
        request_from_other_domain = Float()
        anchor_from_other_domain = Float()
        is_popular = Float()
        long_url = Float()
        age_of_domain = Integer()
        ip_in_url = Integer()
        label = Integer()
    """
    dataset = None

    def __init__(self):
        self.dataset = iter(Phishing())

    def next(self):
        X, Y = next(self.dataset)
        r = Webpage(**X)
        r.label = int(Y)
        return r

    @staticmethod
    def schema():
        return JsonSchema(Webpage)


class Webpage(Record):
    empty_server_form_handler = Float()
    popup_window = Float()
    https = Float()
    request_from_other_domain = Float()
    anchor_from_other_domain = Float()
    is_popular = Float()
    long_url = Float()
    age_of_domain = Integer()
    ip_in_url = Integer()
    label = Integer()



from river.reco.baseline import Baseline
from river.reco.biased_mf import BiasedMF
from river.reco.funk_mf import FunkMF
from river.reco.normal import RandomNormal

__all__ = ["Baseline", "BiasedMF", "FunkMF", "RandomNormal"]

try:
    from .surprise import SurpriseWrapper

    __all__ += ["SurpriseWrapper"]
except ImportError:
    pass



import os
from . import model
from . import function
from . import data
from . import core
from . import metrics
from . import stats

__all__ = ['model', 'function', 'data', 'core', 'metrics', 'stats']


def requirements():
    f = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'requirements.txt'), "r")
    return f.read().splitlines()



from river.neighbors.knn_adwin import KNNADWINClassifier
from river.neighbors.knn_classifier import KNNClassifier
from river.neighbors.knn_regressor import KNNRegressor
from river.neighbors.sam_knn import SAMKNNClassifier

__all__ = ["KNNADWINClassifier", "KNNClassifier", "KNNRegressor", "SAMKNNClassifier"]


from river.linear_model.alma import ALMAClassifier
from river.linear_model.glm import LinearRegression, LogisticRegression, Perceptron
from river.linear_model.pa import PAClassifier, PARegressor
from river.linear_model.softmax import SoftmaxRegression

__all__ = [
    "ALMAClassifier",
    "LinearRegression",
    "LogisticRegression",
    "PAClassifier",
    "PARegressor",
    "Perceptron",
    "SoftmaxRegression",
]


from river.stream.cache import Cache
from river.stream.iter_arff import iter_arff
from river.stream.iter_array import iter_array
from river.stream.iter_csv import iter_csv
from river.stream.iter_libsvm import iter_libsvm
from river.stream.qa import simulate_qa
from river.stream.shuffling import shuffle

__all__ = [
    "Cache",
    "iter_arff",
    "iter_array",
    "iter_csv",
    "iter_libsvm",
    "simulate_qa",
    "shuffle",
]

try:
    from .iter_pandas import iter_pandas

    __all__ += ["iter_pandas"]
except ImportError:
    pass

try:
    from .iter_sklearn import iter_sklearn_dataset

    __all__ += ["iter_sklearn_dataset"]
except ImportError:
    pass

try:
    from .iter_sql import iter_sql

    __all__ += ["iter_sql"]
except ImportError:
    pass

try:
    from .iter_vaex import iter_vaex

    __all__ += ["iter_vaex"]
except ImportError:
    pass


from abc import abstractmethod
from enum import Enum, EnumMeta
from collections import OrderedDict
from six import with_metaclass
import json
import enum
import importlib
pulsar_spec = importlib.util.find_spec("pulsar")

if pulsar_spec is not None:

    from pulsar.schema import Record, Field, Null, Boolean, Integer, Long, Float, Double, Bytes, String, \
        Array, Map, JsonSchema, Schema, BytesSchema, StringSchema, JsonSchema, AvroSchema
else:

    class Schema(object):
        def __init__(self, record_cls, schema_type, schema_definition, schema_name):
            self._record_cls = record_cls

        @abstractmethod
        def encode(self, obj):
            pass

        @abstractmethod
        def decode(self, data):
            pass

        def _validate_object_type(self, obj):
            if not isinstance(obj, self._record_cls):
                raise TypeError('Invalid record obj of type ' + str(type(obj))
                                + ' - expected type is ' + str(self._record_cls))


    class JsonSchema(Schema):

        def __init__(self, record_cls):
            super(JsonSchema, self).__init__(record_cls)

        def _get_serialized_value(self, o):
            if isinstance(o, enum.Enum):
                return o.value
            else:
                return o.__dict__

        def encode(self, obj):
            self._validate_object_type(obj)
            return json.dumps(obj.__dict__, default=self._get_serialized_value, indent=True).encode('utf-8')

        def decode(self, data):
            return self._record_cls(**json.loads(data))


    def _check_record_or_field(x):
        if (type(x) is type and not issubclass(x, Record)) \
                and not isinstance(x, Field):
            raise Exception('Argument ' + x + ' is not a Record or a Field')


    class RecordMeta(type):
        def __new__(metacls, name, parents, dct):
            if name != 'Record':
                # Do not apply this logic to the base class itself
                dct['_fields'] = RecordMeta._get_fields(dct)
                dct['_required'] = False
            return type.__new__(metacls, name, parents, dct)

        @classmethod
        def _get_fields(cls, dct):
            # Build a set of valid fields for this record
            fields = OrderedDict()
            for name, value in dct.items():
                if issubclass(type(value), EnumMeta):
                    # Wrap Python enums
                    value = _Enum(value)
                elif type(value) == RecordMeta:
                    # We expect an instance of a record rather than the class itself
                    value = value()

                if isinstance(value, Record) or isinstance(value, Field):
                    fields[name] = value
            return fields


    class Record(with_metaclass(RecordMeta, object)):

        def __init__(self, *args, **kwargs):
            if args:
                # Only allow keyword args
                raise TypeError('Non-keyword arguments not allowed when initializing Records')

            for k, value in self._fields.items():
                if k in kwargs:
                    # Value was overridden at constructor
                    self.__setattr__(k, kwargs[k])
                elif isinstance(value, Record):
                    # Value is a subrecord
                    self.__setattr__(k, value)
                else:
                    # Set field to default value, without revalidating the default value type
                    super(Record, self).__setattr__(k, value.default())

        @classmethod
        def schema(cls):
            schema = {
                'name': str(cls.__name__),
                'type': 'record',
                'fields': []
            }

            for name in sorted(cls._fields.keys()):
                field = cls._fields[name]
                field_type = field.schema() if field._required else ['null', field.schema()]
                schema['fields'].append({
                    'name': name,
                    'type': field_type
                })
            return schema

        def __setattr__(self, key, value):
            if key not in self._fields:
                raise AttributeError('Cannot set undeclared field ' + key + ' on record')

            # Check that type of value matches the field type
            field = self._fields[key]
            value = field.validate_type(key, value)
            super(Record, self).__setattr__(key, value)

        def __eq__(self, other):
            for field in self._fields:
                if self.__getattribute__(field) != other.__getattribute__(field):
                    return False
            return True

        def __str__(self):
            return str(self.__dict__)

        def type(self):
            return str(self.__class__.__name__)

        def validate_type(self, name, val):
            if not isinstance(val, self.__class__):
                raise TypeError("Invalid type '%s' for sub-record field '%s'. Expected: %s" % (
                    type(val), name, self.__class__))
            return val


    class Field(object):
        def __init__(self, default=None, required=False):
            if default is not None:
                default = self.validate_type('default', default)
            self._default = default
            self._required = required

        @abstractmethod
        def type(self):
            pass

        @abstractmethod
        def python_type(self):
            pass

        def validate_type(self, name, val):
            if type(val) != self.python_type():
                raise TypeError("Invalid type '%s' for field '%s'. Expected: %s" % (type(val), name, self.python_type()))
            return val

        def schema(self):
            # For primitive types, the schema would just be the type itself
            return self.type()

        def default(self):
            return self._default

    # All types


    class Null(Field):
        def type(self):
            return 'null'

        def python_type(self):
            return type(None)

        def validate_type(self, name, val):
            if val is not None:
                raise TypeError('Field ' + name + ' is set to be None')
            return val


    class Boolean(Field):
        def type(self):
            return 'boolean'

        def python_type(self):
            return bool


    class Integer(Field):
        def type(self):
            return 'int'

        def python_type(self):
            return int


    class Long(Field):
        def type(self):
            return 'long'

        def python_type(self):
            return int


    class Float(Field):
        def type(self):
            return 'float'

        def python_type(self):
            return float


    class Double(Field):
        def type(self):
            return 'double'

        def python_type(self):
            return float


    class Bytes(Field):
        def type(self):
            return 'bytes'

        def python_type(self):
            return bytes


    class String(Field):
        def type(self):
            return 'string'

        def python_type(self):
            return str

        def validate_type(self, name, val):
            t = type(val)
            if not (t is str or t.__name__ == 'unicode'):
                raise TypeError("Invalid type '%s' for field '%s'. Expected a string" % (t, name))
            return val

    # Complex types


    class _Enum(Field):
        def __init__(self, enum_type):
            if not issubclass(enum_type, Enum):
                raise Exception(enum_type + " is not a valid Enum type")
            self.enum_type = enum_type
            self.values = {}
            for x in enum_type.__members__.values():
                self.values[x.value] = x
            super(_Enum, self).__init__()

        def type(self):
            return 'enum'

        def python_type(self):
            return self.enum_type

        def validate_type(self, name, val):
            if type(val) is str:
                # The enum was passed as a string, we need to check it against the possible values
                if val in self.enum_type.__members__:
                    return self.enum_type.__members__[val]
                else:
                    raise TypeError(
                        "Invalid enum value '%s' for field '%s'. Expected: %s" % (val, name, self.enum_type.__members__.keys()))
            elif type(val) is int:
                # The enum was passed as an int, we need to check it against the possible values
                if val in self.values:
                    return self.values[val]
                else:
                    raise TypeError(
                        "Invalid enum value '%s' for field '%s'. Expected: %s" % (val, name, self.values.keys()))
            elif type(val) != self.python_type():
                raise TypeError("Invalid type '%s' for field '%s'. Expected: %s" % (type(val), name, self.python_type()))
            else:
                return val

        def schema(self):
            return {
                'type': self.type(),
                'name': self.enum_type.__name__,
                'symbols': [x.name for x in self.enum_type]
            }


    class Array(Field):
        def __init__(self, array_type):
            _check_record_or_field(array_type)
            self.array_type = array_type
            super(Array, self).__init__()

        def type(self):
            return 'array'

        def python_type(self):
            return list

        def validate_type(self, name, val):
            super(Array, self).validate_type(name, val)

            for x in val:
                if type(x) != self.array_type.python_type():
                    raise TypeError('Array field ' + name + ' items should all be of type '
                                    + self.array_type.python_type())
            return val

        def schema(self):
            return {
                'type': self.type(),
                'items': self.array_type.schema() if isinstance(self.array_type, Record)
                    else self.array_type.type()
            }


    class Map(Field):
        def __init__(self, value_type):
            _check_record_or_field(value_type)
            self.value_type = value_type
            super(Map, self).__init__()

        def type(self):
            return 'map'

        def python_type(self):
            return dict

        def validate_type(self, name, val):
            super(Map, self).validate_type(name, val)

            for k, v in val.items():
                if type(k) != str:
                    raise TypeError('Map keys for field ' + name + '  should all be strings')
                if type(v) != self.value_type.python_type():
                    raise TypeError('Map values for field ' + name + ' should all be of type '
                                    + self.value_type.python_type())

            return val

        def schema(self):
            return {
                'type': self.type(),
                'values': self.value_type.schema() if isinstance(self.value_type, Record)
                    else self.value_type.type()
            }



import pickle
import codecs

__all__ = [
    "storage"
]


class Storage:
    container = None

    def __init__(self, driver='local'):
        if driver == 'local':
            from cloudstorage.drivers.local import LocalDriver
            storage = LocalDriver(key='./models')
        elif driver == 'google':
            from cloudstorage.drivers.amazon import S3Driver
            storage = S3Driver(key='<my-aws-access-key-id>', secret='<my-aws-secret-access-key>')
        elif driver == 'amazon':
            from cloudstorage.drivers.google import GoogleStorageDriver
            storage = GoogleStorageDriver(key='<my-aws-access-key-id>')
        else:
            raise Exception(f"Invalid driver specified: {driver}")

        self.container = storage.create_container('saved')

    def put(self, object_name, object):
        return self.container.upload_blob(codecs.encode(pickle.dumps(object), "base64").decode(), blob_name=object_name)

    def get(self, object_name):
        return pickle.loads(codecs.decode(self.container.get_blob(object_name).encode(), "base64"))


storage = Storage()



from river.facto.ffm import FFMClassifier, FFMRegressor
from river.facto.fm import FMClassifier, FMRegressor
from river.facto.fwfm import FwFMClassifier, FwFMRegressor
from river.facto.hofm import HOFMClassifier, HOFMRegressor

__all__ = [
    "FFMClassifier",
    "FFMRegressor",
    "FMClassifier",
    "FMRegressor",
    "FwFMClassifier",
    "FwFMRegressor",
    "HOFMClassifier",
    "HOFMRegressor",
]


from river.stats.auto_corr import AutoCorr
from river.stats.base import Bivariate, Univariate
from river.stats.count import Count
from river.stats.cov import Cov, RollingCov
from river.stats.entropy import Entropy
from river.stats.ewmean import EWMean
from river.stats.ewvar import EWVar
from river.stats.iqr import IQR, RollingIQR
from river.stats.kurtosis import Kurtosis
from river.stats.link import Link
from river.stats.maximum import AbsMax, Max, RollingAbsMax, RollingMax
from river.stats.mean import BayesianMean, Mean, RollingMean
from river.stats.minimum import Min, RollingMin
from river.stats.mode import Mode, RollingMode
from river.stats.n_unique import NUnique
from river.stats.pearson import PearsonCorr, RollingPearsonCorr
from river.stats.ptp import PeakToPeak, RollingPeakToPeak
from river.stats.quantile import Quantile, RollingQuantile
from river.stats.sem import SEM, RollingSEM
from river.stats.shift import Shift
from river.stats.skew import Skew
from river.stats.summing import RollingSum, Sum
from river.stats.var import RollingVar, Var

__all__ = [
    "AbsMax",
    "AutoCorr",
    "BayesianMean",
    "Bivariate",
    "Count",
    "Cov",
    "Entropy",
    "EWMean",
    "EWVar",
    "IQR",
    "Kurtosis",
    "Link",
    "Max",
    "Mean",
    "Min",
    "Mode",
    "NUnique",
    "PeakToPeak",
    "PearsonCorr",
    "Quantile",
    "RollingAbsMax",
    "RollingCov",
    "RollingIQR",
    "RollingMax",
    "RollingMean",
    "RollingMin",
    "RollingMode",
    "RollingPeakToPeak",
    "RollingPearsonCorr",
    "RollingQuantile",
    "RollingSEM",
    "RollingSum",
    "RollingVar",
    "SEM",
    "Shift",
    "Skew",
    "Sum",
    "Univariate",
    "Var",
]



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Integer
from .scikit_multiflow_classes import SKStream, check_random_state
import numpy as np


class LEDGeneratorDataset(Stream):
    """
    A generator for data regarding a digit displayed on a seven-segment LED display with the ability to add noise.

    Each record is an instance of Sine

    class LED(Record):
        one = Integer()
        two = Integer()
        three = Integer()
        four = Integer()
        five = Integer()
        six = Integer()
        seven = Integer()
        eight = Integer()
        nine = Integer()
        ten = Integer()
        eleven = Integer()
        twelve = Integer()
        thirteen = Integer()
        fourteen = Integer()
        fifteen = Integer()
        sixteen = Integer()
        seventeen = Integer()
        eighteen = Integer()
        nineteen = Integer()
        twenty = Integer()
        twentyone = Integer()
        twentytwo = Integer()
        twentythree = Integer()
        twentyfour = Integer()
        label = Integer()

    LED stream generator.

    This data source originates from the CART book [1]_. An implementation
    in C was donated to the UCI [2]_ machine learning repository by David Aha.
    The goal is to predict the digit displayed on a seven-segment LED display,
    where each attribute has a 10% chance of being inverted. It has an optimal
    Bayes classification rate of 74%. The particular configuration of the
    generator used for experiments ( LED ) produces 24 binary attributes,
    17 of which are irrelevant.

    Parameters
    ----------
    random_state: int, RandomState instance or None, optional (default=None)
       If int, random_state is the seed used by the random number generator;
       If RandomState instance, random_state is the random number generator;
       If None, the random number generator is the RandomState instance used
       by `np.random`.

    noise_percentage: float (Default: 0.0)
      The probability that noise will happen in the generation. At each
      new sample generated, a random probability is generated, and if that
      probability is equal or less than the noise_percentage, the selected
      data  will be switched

    has_noise: bool (Default: False)
       Adds 17 non relevant attributes to the stream.

    References
    ----------

    .. [1] Leo Breiman, Jerome Friedman, R. Olshen, and Charles J. Stone.
           Classification and Regression Trees. Wadsworth and Brooks,
           Monterey, CA,1984.

    .. [2] A. Asuncion and D. J. Newman. UCI Machine Learning Repository
          [http://www.ics.uci.edu/∼mlearn/mlrepository.html].
          University of California, Irvine, School of Information and
          Computer Sciences,2007.

    """
    dataset = None

    def __init__(self, *args, **kwargs):
        self.dataset = LEDGenerator(*args, **kwargs)

    def next(self):
        features, label = self.dataset.next_sample()

        dict = {'label': label[0].item()}

        keys = list(LED._fields.keys())

        index = 0
        for element in features[0]:
            dict[keys[index]] = int(element.item())
            index += 1

        return LED(**dict)

    @staticmethod
    def schema():
        return JsonSchema(LED)


class LED(Record):
    one = Integer()
    two = Integer()
    three = Integer()
    four = Integer()
    five = Integer()
    six = Integer()
    seven = Integer()
    label = Integer()
    

class LEDGenerator(SKStream):
    """ LED stream generator.

    This data source originates from the CART book [1]_. An implementation
    in C was donated to the UCI [2]_ machine learning repository by David Aha.
    The goal is to predict the digit displayed on a seven-segment LED display,
    where each attribute has a 10% chance of being inverted. It has an optimal
    Bayes classification rate of 74%. The particular configuration of the
    generator used for experiments ( LED ) produces 24 binary attributes,
    17 of which are irrelevant.

    Parameters
    ----------
    random_state: int, RandomState instance or None, optional (default=None)
       If int, random_state is the seed used by the random number generator;
       If RandomState instance, random_state is the random number generator;
       If None, the random number generator is the RandomState instance used
       by `np.random`.

    noise_percentage: float (Default: 0.0)
      The probability that noise will happen in the generation. At each
      new sample generated, a random probability is generated, and if that
      probability is equal or less than the noise_percentage, the selected
      data  will be switched

    has_noise: bool (Default: False)
       Adds 17 non relevant attributes to the stream.

    References
    ----------

    .. [1] Leo Breiman, Jerome Friedman, R. Olshen, and Charles J. Stone.
           Classification and Regression Trees. Wadsworth and Brooks,
           Monterey, CA,1984.

    .. [2] A. Asuncion and D. J. Newman. UCI Machine Learning Repository
          [http://www.ics.uci.edu/∼mlearn/mlrepository.html].
          University of California, Irvine, School of Information and
          Computer Sciences,2007.


    Examples
    --------
    >>> # Imports
    >>> from skmultiflow.data.led_generator import LEDGenerator
    >>> # Setting up the stream
    >>> stream = LEDGenerator(random_state = 112, noise_percentage = 0.28, has_noise= True)
    >>> # Retrieving one sample
    >>> stream.next_sample()
    (array([[0., 1., 1., 1., 0., 0., 0., 0., 1., 0., 0., 0., 1., 0., 1., 1.,
      1., 0., 0., 1., 1., 0., 1., 1.]]), array([4]))

    >>> # Retrieving 10 samples
    >>> stream.next_sample(10)
    (array([[0., 0., 1., 0., 0., 0., 0., 1., 0., 0., 1., 1., 0., 0., 0., 0.,
      1., 1., 1., 0., 0., 0., 1., 1.],
     [1., 1., 1., 0., 1., 0., 1., 1., 1., 0., 1., 0., 0., 0., 1., 1.,
      1., 1., 0., 0., 1., 0., 1., 0.],
     [0., 1., 1., 0., 0., 1., 1., 1., 0., 0., 0., 0., 1., 0., 0., 0.,
      0., 1., 0., 1., 1., 1., 1., 1.],
     [1., 1., 0., 0., 0., 1., 1., 1., 0., 1., 1., 0., 1., 1., 0., 0.,
      1., 1., 1., 0., 0., 0., 1., 0.],
     [1., 1., 1., 0., 0., 1., 0., 0., 1., 1., 0., 1., 1., 0., 1., 0.,
      0., 0., 1., 0., 1., 0., 0., 0.],
     [0., 1., 1., 0., 0., 1., 0., 0., 1., 1., 0., 1., 0., 1., 1., 1.,
      0., 0., 1., 0., 1., 1., 0., 0.],
     [0., 0., 0., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0., 1., 0.,
      1., 1., 1., 0., 1., 0., 0., 1.],
     [0., 0., 0., 0., 0., 1., 0., 1., 1., 1., 0., 0., 0., 0., 0., 1.,
      1., 1., 1., 1., 0., 1., 1., 1.],
     [1., 1., 1., 0., 0., 1., 0., 1., 1., 1., 0., 1., 1., 1., 1., 1.,
      0., 1., 1., 0., 0., 0., 0., 1.],
     [1., 1., 1., 0., 0., 1., 1., 0., 0., 0., 0., 0., 1., 0., 0., 0.,
      1., 1., 0., 0., 0., 0., 1., 0.]]),
     array([1, 0, 7, 9, 7, 1, 3, 1, 4, 1]))

     >>> stream.n_remaining_samples()
     -1
     >>> stream.has_more_samples()
     True

    """
    _NUM_BASE_ATTRIBUTES = 7
    _TOTAL_ATTRIBUTES_INCLUDING_NOISE = 24
    _ORIGINAL_INSTANCES = np.array([[1, 1, 1, 0, 1, 1, 1],
                                    [0, 0, 1, 0, 0, 1, 0],
                                    [1, 0, 1, 1, 1, 0, 1],
                                    [1, 0, 1, 1, 0, 1, 1],
                                    [0, 1, 1, 1, 0, 1, 0],
                                    [1, 1, 0, 1, 0, 1, 1],
                                    [1, 1, 0, 1, 1, 1, 1],
                                    [1, 0, 1, 0, 0, 1, 0],
                                    [1, 1, 1, 1, 1, 1, 1],
                                    [1, 1, 1, 1, 0, 1, 1]])

    def __init__(self, random_state=None, noise_percentage=0.0, has_noise=False):
        super().__init__()
        self.random_state = random_state
        self._random_state = None  # This is the actual random_state object used internally
        self.noise_percentage = noise_percentage
        self.n_cat_features = self._NUM_BASE_ATTRIBUTES
        self.n_features = self.n_cat_features
        self.has_noise = has_noise
        self.n_targets = 1
        self.n_classes = 10
        self.name = "Led Generator"

        if self.has_noise:
            self.n_cat_features = self._TOTAL_ATTRIBUTES_INCLUDING_NOISE
        else:
            self.n_cat_features = self._NUM_BASE_ATTRIBUTES

        self.n_features = self.n_cat_features
        self.feature_names = ["att_num_" + str(i) for i in range(self.n_cat_features)]
        self.target_values = [i for i in range(self.n_classes)]

        self._prepare_for_use()

    @property
    def noise_percentage(self):
        """ Retrieve the value of the option: Noise percentage

        Returns
        -------
        float
            The value of the noise percentage
        """
        return self._noise_percentage

    @noise_percentage.setter
    def noise_percentage(self, noise_percentage):
        """ Set the value of the option: Noise percentage

        Parameters
        ----------
        noise_percentage: float (0.0..1.0)

        """
        if (0.0 <= noise_percentage) and (noise_percentage <= 1.0):
            self._noise_percentage = noise_percentage
        else:
            raise ValueError("noise percentage should be in [0.0..1.0], and {} was passed".format(
                noise_percentage))

    @property
    def has_noise(self):
        """ Retrieve the value of the option: add noise.

        Returns
        -------
        Boolean
            True is the noise is added.
        """
        return self._has_noise

    @has_noise.setter
    def has_noise(self, has_noise):
        """ Set the value of the option: add noise.

        Parameters
        ----------
        has_noise: Boolean

        """
        if isinstance(has_noise, bool):
            self._has_noise = has_noise
        else:
            raise ValueError("has_noise should be boolean, and {} was passed".format(has_noise))

    def _prepare_for_use(self):
        self._random_state = check_random_state(self.random_state)

    def next_sample(self, batch_size=1):
        """ Returns next sample from the stream.

        An instance is generated based on the parameters passed. If noise
        is included the total number of attributes will be 24, if it's not
        included there will be 7 attributes.

        Parameters
        ----------
        batch_size: int (optional, default=1)
            The number of samples to return.

        Returns
        -------
        tuple or tuple list
            Return a tuple with the features matrix
            for the batch_size samples that were requested.

        """

        data = np.zeros([batch_size, self.n_features + 1])
        target = np.zeros(batch_size, dtype=int)

        for j in range(batch_size):
            self.sample_idx += 1
            selected = self._random_state.randint(self.n_classes)
            target[j] = selected

            for i in range(self._NUM_BASE_ATTRIBUTES):
                if (0.01 + self._random_state.rand()) <= self.noise_percentage:
                    data[j, i] = 1 if (self._ORIGINAL_INSTANCES[selected, i] == 0) else 0
                else:
                    data[j, i] = self._ORIGINAL_INSTANCES[selected, i]

            if self.has_noise:
                for i in range(self._NUM_BASE_ATTRIBUTES, self._TOTAL_ATTRIBUTES_INCLUDING_NOISE):
                    data[j, i] = self._random_state.randint(2)

        self.current_sample_x = data[:, :self.n_features]
        self.current_sample_y = target
        return self.current_sample_x, self.current_sample_y

    def get_data_info(self):
        return "Led Generator - {} features".format(self.n_features)



from river.evaluate.progressive_validation import progressive_val_score
from river.evaluate.tracks import Track, load_binary_clf_tracks

__all__ = ["load_binary_clf_tracks", "progressive_val_score", "Track"]


from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Integer, Float, String
from river.datasets import MovieLens100K


class MovieRatingDataset(Stream):
    """
    This dataset contains 100,000 movie ratings from a variety of individuals.

    Each record is an instance of MovieRating

    class MovieRating(Record):
        user_id = Integer()
        item_id = Integer()
        timestamp = Integer()
        title = String()
        release_date = Integer()
        genres = String()
        user_age = Integer()
        user_gender = String()
        user_occupation = String()
        user_zip_code = String()
        user_movie_rating = Float()
    """
    dataset = None

    def __init__(self):
        self.dataset = iter(MovieLens100K())

    def next(self):
        X, Y = next(self.dataset)
        return MovieRating(user_id = int(X['user']), item_id = int(X['item']), timestamp=X['timestamp'],
                           title=X['title'], release_date=X['release_date'], genres=X['genres'], user_age=int(X['age']),
                           user_gender=X['gender'], occupation=X['occupation'], user_zip_code=X['zip_code'],
                           user_movie_rating=Y)

    @staticmethod
    def schema():
        return JsonSchema(MovieRating)


class MovieRating(Record):
    user_id = Integer()
    item_id = Integer()
    timestamp = Integer()
    title = String()
    release_date = Integer()
    genres = String()
    user_age = Integer()
    user_gender = String()
    user_occupation = String()
    user_zip_code = String()
    user_movie_rating = Float()



import code


# local=locals() or local=globals()
def interact(banner=None, local=locals(), exitmsg=None):
    return code.interact(banner='Start PandioML Interactive Session' + (' | ' + banner if banner is not None else ''),
                         readfunc=None, local=local, exitmsg='Exit PandioML Interactive Session' +
                                                             (' | ' + exitmsg if exitmsg is not None else ''))



import numpy as np
from abc import ABCMeta, abstractmethod
import warnings

__all__ = ['BaseEstimator', 'BaseSKMObject', 'SKStream', 'check_random_state']


class BaseEstimator:
    """Base Estimator class for compatibility with scikit-learn.

    Notes
    -----
    * All estimators should specify all the parameters that can be set
      at the class level in their ``__init__`` as explicit keyword
      arguments (no ``*args`` or ``**kwargs``).
    * Taken from sklearn for compatibility.
    """

    @classmethod
    def _get_param_names(cls):
        """Get parameter names for the estimator"""
        # fetch the constructor or the original constructor before
        # deprecation wrapping if any
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError("scikit-multiflow estimators should always "
                                   "specify their parameters in the signature"
                                   " of their __init__ (no varargs)."
                                   " %s with constructor %s doesn't "
                                   " follow this convention."
                                   % (cls, init_signature))
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def get_params(self, deep=True):
        """Get parameters for this estimator.

        Parameters
        ----------
        deep : boolean, optional
            If True, will return the parameters for this estimator and
            contained subobjects that are estimators.

        Returns
        -------
        params : mapping of string to any
            Parameter names mapped to their values.
        """
        out = dict()
        for key in self._get_param_names():
            value = getattr(self, key, None)
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    def set_params(self, **params):
        """Set the parameters of this estimator.

        The method works on simple estimators as well as on nested objects
        (such as pipelines). The latter have parameters of the form
        ``<component>__<parameter>`` so that it's possible to update each
        component of a nested object.

        Returns
        -------
        self
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        nested_params = defaultdict(dict)  # grouped by prefix
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError('Invalid parameter %s for estimator %s. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.' %
                                 (key, self))

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)

        return self

    def __repr__(self, N_CHAR_MAX=700):
        # N_CHAR_MAX is the (approximate) maximum number of non-blank
        # characters to render. We pass it as an optional parameter to ease
        # the tests.

        from ..utils._pprint import _EstimatorPrettyPrinter

        N_MAX_ELEMENTS_TO_SHOW = 30  # number of elements to show in sequences

        # use ellipsis for sequences with a lot of elements
        pp = _EstimatorPrettyPrinter(
            compact=True, indent=1, indent_at_name=True,
            n_max_elements_to_show=N_MAX_ELEMENTS_TO_SHOW)

        repr_ = pp.pformat(self)

        # Use bruteforce ellipsis when there are a lot of non-blank characters
        n_nonblank = len(''.join(repr_.split()))
        if n_nonblank > N_CHAR_MAX:
            lim = N_CHAR_MAX // 2  # apprx number of chars to keep on both ends
            regex = r'^(\s*\S){%d}' % lim
            # The regex '^(\s*\S){%d}' % n
            # matches from the start of the string until the nth non-blank
            # character:
            # - ^ matches the start of string
            # - (pattern){n} matches n repetitions of pattern
            # - \s*\S matches a non-blank char following zero or more blanks
            left_lim = re.match(regex, repr_).end()
            right_lim = re.match(regex, repr_[::-1]).end()

            if '\n' in repr_[left_lim:-right_lim]:
                # The left side and right side aren't on the same line.
                # To avoid weird cuts, e.g.:
                # categoric...ore',
                # we need to start the right side with an appropriate newline
                # character so that it renders properly as:
                # categoric...
                # handle_unknown='ignore',
                # so we add [^\n]*\n which matches until the next \n
                regex += r'[^\n]*\n'
                right_lim = re.match(regex, repr_[::-1]).end()

            ellipsis = '...'
            if left_lim + len(ellipsis) < len(repr_) - right_lim:
                # Only add ellipsis if it results in a shorter repr
                repr_ = repr_[:left_lim] + '...' + repr_[-right_lim:]

        return repr_

    def __getstate__(self):
        try:
            state = super().__getstate__()
        except AttributeError:
            state = self.__dict__.copy()

        if type(self).__module__.startswith('skmultiflow.'):
            return dict(state.items(), _skmultiflow_version=__version__)
        else:
            return state

    def __setstate__(self, state):
        if type(self).__module__.startswith('skmultiflow.'):
            pickle_version = state.pop("_skmultiflow_version", "pre-0.18")
            if pickle_version != __version__:
                warnings.warn(
                    "Trying to unpickle estimator {0} from version {1} when "
                    "using version {2}. This might lead to breaking code or "
                    "invalid results. Use at your own risk.".format(
                        self.__class__.__name__, pickle_version, __version__),
                    UserWarning)
        try:
            super().__setstate__(state)
        except AttributeError:
            self.__dict__.update(state)

    def _get_tags(self):
        collected_tags = {}
        for base_class in inspect.getmro(self.__class__):
            if (hasattr(base_class, '_more_tags') and base_class != self.__class__):
                more_tags = base_class._more_tags(self)
                collected_tags = _update_if_consistent(collected_tags,
                                                       more_tags)
        if hasattr(self, '_more_tags'):
            more_tags = self._more_tags()
            collected_tags = _update_if_consistent(collected_tags, more_tags)
        tags = _DEFAULT_TAGS.copy()
        tags.update(collected_tags)
        return tags


class BaseSKMObject(BaseEstimator):
    """Base class for most objects in scikit-multiflow

        Notes
        -----
        This class provides additional functionality not available in the base estimator
        from scikit-learn
    """
    def reset(self):
        """ Resets the estimator to its initial state.

        Returns
        -------
        self

        """
        # non-optimized default implementation; override if a better
        # method is possible for a given object
        command = ''.join([line.strip() for line in self.__repr__().split()])
        command = command.replace(str(self.__class__.__name__), 'self.__init__')
        exec(command)

    def get_info(self):
        """ Collects and returns the information about the configuration of the estimator

        Returns
        -------
        string
            Configuration of the estimator.
        """
        return self.__repr__()


class SKStream(BaseSKMObject, metaclass=ABCMeta):
    """ Base Stream class.

    This abstract class defines the minimum requirements of a stream,
    so that it can work along other modules in scikit-multiflow.

    Raises
    ------
    NotImplementedError: This is an abstract class.

    """
    _estimator_type = 'stream'

    def __init__(self):
        self.n_samples = 0
        self.n_targets = 0
        self.n_features = 0
        self.n_num_features = 0
        self.n_cat_features = 0
        self.n_classes = 0
        self.cat_features_idx = []
        self.current_sample_x = None
        self.current_sample_y = None
        self.sample_idx = 0
        self.feature_names = None
        self.target_names = None
        self.target_values = None
        self.name = None

    @property
    def n_features(self):
        """ Retrieve the number of features.

        Returns
        -------
        int
            The total number of features.

        """
        return self._n_features

    @n_features.setter
    def n_features(self, n_features):
        """ Set the number of features

        """
        self._n_features = n_features

    @property
    def n_cat_features(self):
        """ Retrieve the number of integer features.

        Returns
        -------
        int
            The number of integer features in the stream.

        """
        return self._n_cat_features

    @n_cat_features.setter
    def n_cat_features(self, n_cat_features):
        """ Set the number of integer features

        Parameters
        ----------
        n_cat_features: int
        """
        self._n_cat_features = n_cat_features

    @property
    def n_num_features(self):
        """ Retrieve the number of numerical features.

        Returns
        -------
        int
            The number of numerical features in the stream.

        """
        return self._n_num_features

    @n_num_features.setter
    def n_num_features(self, n_num_features):
        """ Set the number of numerical features

        Parameters
        ----------
        n_num_features: int

        """
        self._n_num_features = n_num_features

    @property
    def n_targets(self):
        """ Retrieve the number of targets

        Returns
        -------
        int
            the number of targets in the stream.
        """
        return self._target_idx

    @n_targets.setter
    def n_targets(self, n_targets):
        """ Set the number of targets

        Parameters
        ----------
        n_targets: int
        """
        self._target_idx = n_targets

    @property
    def target_values(self):
        """ Retrieve all target_values in the stream for each target.

        Returns
        -------
        list
            list of lists of all target_values for each target
        """
        return self._target_values

    @target_values.setter
    def target_values(self, target_values):
        """ Set the list for all target_values in the stream.

        Parameters
        ----------
        target_values
        """
        self._target_values = target_values

    @property
    def feature_names(self):
        """ Retrieve the names of the features.

        Returns
        -------
        list
            names of the features
        """
        return self._feature_names

    @feature_names.setter
    def feature_names(self, feature_names):
        """ Set the name of the features in the stream.

        Parameters
        ----------
        feature_names: list
        """
        self._feature_names = feature_names

    @property
    def target_names(self):
        """ Retrieve the names of the targets

        Returns
        -------
        list
            the names of the targets in the stream.
        """
        return self._target_names

    @target_names.setter
    def target_names(self, target_names):
        """ Set the names of the targets in the stream.

        Parameters
        ----------
        target_names: list

        """
        self._target_names = target_names

    @staticmethod
    def prepare_for_use():  # pragma: no cover
        """ Prepare the stream for use.

        Deprecated in v0.5.0 and will be removed in v0.7.0

        """
        warnings.warn(
            "'prepare_for_use' has been deprecated in v0.5.0 and will be removed in v0.7.0.\n"
            "New instances of the Stream class are now ready to use after instantiation.",
            category=FutureWarning)

    @abstractmethod
    def _prepare_for_use(self):
        raise NotImplementedError

    @abstractmethod
    def next_sample(self, batch_size=1):
        """ Returns next sample from the stream.

        Parameters
        ----------
        batch_size: int (optional, default=1)
            The number of samples to return.

        Returns
        -------
        tuple or tuple list
            A numpy.ndarray of shape (batch_size, n_features) and an array-like of size
            n_targets, representing the next batch_size samples.

        """
        raise NotImplementedError

    def last_sample(self):
        """ Retrieves last `batch_size` samples in the stream.

        Returns
        -------
        tuple or tuple list
            A numpy.ndarray of shape (batch_size, n_features) and an array-like of shape
            (batch_size, n_targets), representing the next batch_size samples.

        """
        return self.current_sample_x, self.current_sample_y

    def is_restartable(self):
        """
        Determine if the stream is restartable.

        Returns
        -------
        Bool
            True if stream is restartable.

        """
        return True

    def restart(self):
        """  Restart the stream. """
        self.current_sample_x = None
        self.current_sample_y = None
        self.sample_idx = 0
        self._prepare_for_use()

    def n_remaining_samples(self):
        """ Returns the estimated number of remaining samples.

        Returns
        -------
        int
            Remaining number of samples. -1 if infinite (e.g. generator)

        """
        return -1

    def has_more_samples(self):
        """
        Checks if stream has more samples.

        Returns
        -------
        Boolean
            True if stream has more samples.
        """
        return True

    def get_data_info(self):
        """ Retrieves minimum information from the stream

        Used by evaluator methods to id the stream.

        The default format is: 'Stream name - n_targets, n_classes, n_features'.

        Returns
        -------
        string
            Stream data information

        """
        return self.name + " - {} target(s), {} classes, {} features".format(self.n_targets,
                                                                             self.n_classes,
                                                                             self.n_features)


def check_random_state(seed):
    """Turn seed into a np.random.RandomState instance.

    Parameters
    ----------
    seed : None | int | instance of RandomState
        If seed is None, return the RandomState singleton used by np.random.
        If seed is an int, return a new RandomState instance seeded with seed.
        If seed is already a RandomState instance, return it.
        Otherwise raise ValueError.

    Notes
    -----
    Code from sklearn

    """
    if seed is None or seed is np.random:
        return np.random.mtrand._rand
    if isinstance(seed, (numbers.Integral, np.integer)):
        return np.random.RandomState(seed)
    if isinstance(seed, np.random.RandomState):
        return seed
    raise ValueError('{} cannot be used to seed a numpy.random.RandomState instance'.format(seed))



from abc import ABCMeta, abstractmethod, abstractproperty
import re
import inspect


class Stream(object, metaclass=ABCMeta):
    @abstractmethod
    def __init__(self):
        raise NotImplementedError

    @property
    @abstractmethod
    def dataset(self):
        raise NotImplementedError

    @abstractmethod
    def next(self):
        raise NotImplementedError

    @abstractmethod
    def schema(self):
        raise NotImplementedError

    @property
    def desc(self):
        """Return the description from the docstring."""
        desc = re.split(pattern=r"\w+\n\s{4}\-{3,}", string=self.__doc__, maxsplit=0)[0]
        return inspect.cleandoc(desc)

    def __repr__(self):
        out = f"{self.desc}"
        if hasattr(self, 'data'):
            out += "\n\n" + self.data.__repr__()

        return out

    def __iter__(self):
        return self.dataset



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Integer, String, Boolean, Double, Float
from river.datasets import Restaurants
from datetime import datetime


class RestaurantVisitorsDataset(Stream):
    """
    This dataset contains 252,108 records over a 16 week period to 829 Japanese Restaurants.

    Each record is an instance of RestaurantDay

    class RestaurantDay(Record):
        store_id = String()
        timestamp = Float()
        is_holiday = Boolean()
        genre_name = String()
        area_name = String()
        latitude = Double()
        longitude = Double()
        visitors = Integer()
    """
    dataset = None

    def __init__(self):
        self.dataset = iter(Restaurants())

    def next(self):
        X, Y = next(self.dataset)
        return RestaurantDay(store_id=X['store_id'], timestamp=datetime.timestamp(X['date']),
                             is_holiday=X['is_holiday'], genre_name=X['genre_name'], area_name=X['area_name'],
                             latitude=X['latitude'], longitude=X['longitude'], visitors=Y)

    @staticmethod
    def schema():
        return JsonSchema(RestaurantDay)


class RestaurantDay(Record):
    store_id = String()
    timestamp = Float()
    is_holiday = Boolean()
    genre_name = String()
    area_name = String()
    latitude = Double()
    longitude = Double()
    visitors = Integer()



from river.expert.bandit import EpsilonGreedyRegressor, UCBRegressor
from river.expert.ewa import EWARegressor
from river.expert.sh import SuccessiveHalvingClassifier, SuccessiveHalvingRegressor
from river.expert.stacking import StackingClassifier

__all__ = [
    "EpsilonGreedyRegressor",
    "EWARegressor",
    "SuccessiveHalvingClassifier",
    "SuccessiveHalvingRegressor",
    "StackingClassifier",
    "UCBRegressor",
]


from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Float, Integer
from .scikit_multiflow_classes import SKStream, check_random_state
import numpy as np


class AgrawalGeneratorDataset(Stream):
    """
    A generator for data regarding home loan applications with the ability to balance and add noise.

    Each record is an instance of LoanApplication

    class LoanApplication(Record):
        salary = Float()
        commission = Float()
        age = Float()
        education_level = Integer()
        car = Float()
        zipcode = Float()
        house_value = Float()
        house_owned_years = Integer()
        loan_amount = Float()
        label = Integer()
    """

    dataset = None

    def __init__(self, *args, **kwargs):
        self.dataset = AGRAWAL(*args, **kwargs)

    def next(self):
        features, label = self.dataset.next_sample()

        dict = {'label': label[0].item()}

        index = 0
        for element in features[0]:
            dict[self.dataset.feature_names[index]] = element.item()
            index += 1

        return LoanApplication(**dict)

    @staticmethod
    def schema():
        return JsonSchema(LoanApplication)


class LoanApplication(Record):
    salary = Float()
    commission = Float()
    age = Float()
    education_level = Integer()
    car = Float()
    zipcode = Float()
    house_value = Float()
    house_owned_years = Integer()
    loan_amount = Float()
    label = Integer()


class AGRAWAL(SKStream):
    """ Agrawal stream generator.

    The generator was introduced by Agrawal et al. in [1]_, and was common source
    of data for early work on scaling up decision tree learners.
    The generator produces a stream containing nine features, six numeric and
    three categorical.
    There are ten functions defined for generating binary class labels from the
    features. Presumably these determine whether the loan should be approved.
    The features and functions are listed in the original paper [1]_.

    .. list-table::
       :widths: 25 25 50
       :header-rows: 1

       * - feature name
         - feature description
         - values
       * - salary
         - the salary
         - uniformly distributed from 20k to 150k
       * - commission
         - the commission
         - if (salary < 75k) then 0 else uniformly distributed from 10k to 75k
       * - age
         - the age
         - uniformly distributed from 20 to 80
       * - elevel
         - the education level
         - uniformly chosen from 0 to 4
       * - car
         - car maker
         - uniformly chosen from 1 to 20
       * - zipcode
         - zip code of the town
         - uniformly chosen from 0 to 8
       * - hvalue
         - value of the house
         - uniformly distributed from 50k x zipcode to 100k x zipcode
       * - hyears
         - years house owned
         - uniformly distributed from 1 to 30
       * - loan
         - total loan amount
         - uniformly distributed from 0 to 500k

    Parameters
    ----------
    classification_function: int (Default=0)
        Which of the four classification functions to use for the generation.
        The value can vary from 0 to 9.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    balance_classes: bool (Default: False)
        Whether to balance classes or not. If balanced, the class
        distribution will converge to a uniform distribution.

    perturbation: float (Default: 0.0)
        The probability that noise will happen in the generation. At each
        new sample generated, the sample with will perturbed by the amount of
        perturbation.
        Values go from 0.0 to 1.0.

    References
    ----------
    .. [1] Rakesh Agrawal, Tomasz Imielinksi, and Arun Swami. "Database
       Mining: A Performance Perspective", IEEE Transactions on Knowledge and
       Data Engineering, 5(6), December 1993.

    """

    def __init__(self, classification_function=0, random_state=None, balance_classes=False,
                 perturbation=0.0):
        super().__init__()

        # Classification functions to use
        self._classification_functions = [self._classification_function_zero,
                                          self._classification_function_one,
                                          self._classification_function_two,
                                          self._classification_function_three,
                                          self._classification_function_four,
                                          self._classification_function_five,
                                          self._classification_function_six,
                                          self._classification_function_seven,
                                          self._classification_function_eight,
                                          self._classification_function_nine]
        self.classification_function = classification_function
        self.random_state = random_state
        self.balance_classes = balance_classes
        self.perturbation = perturbation
        self.n_num_features = 6
        self.n_cat_features = 3
        self.n_features = self.n_num_features + self.n_cat_features
        self.n_classes = 2
        self.n_targets = 1
        self._random_state = None  # This is the actual random_state object used internally
        self._next_class_should_be_zero = False
        self.name = "AGRAWAL Generator"

        self.target_names = ["target"]
        self.feature_names = ["salary", "commission", "age", "elevel", "car", "zipcode", "hvalue",
                              "hyears", "loan"]
        self.target_values = [i for i in range(self.n_classes)]

        self._prepare_for_use()

    @property
    def classification_function(self):
        """ Retrieve the index of the current classification function.

        Returns
        -------
        int
            index of the classification function, from 0 to 9
        """
        return self._classification_function

    @classification_function.setter
    def classification_function(self, classification_function_idx):
        """ Set the index of the current classification function.

        Parameters
        ----------
        classification_function_idx: int
            from 0 to 9
        """
        if classification_function_idx in range(10):
            self._classification_function = classification_function_idx
        else:
            raise ValueError("classification_function takes values from 0 to 9,"
                             " and {} was passed".format(classification_function_idx))

    @property
    def balance_classes(self):
        """ Retrieve the value of the option: Balance classes

        Returns
        -------
        Boolean
            True is the classes are balanced
        """
        return self._balance_classes

    @balance_classes.setter
    def balance_classes(self, balance_classes):
        """ Set the value of the option: Balance classes.

        Parameters
        ----------
        balance_classes: Boolean

        """
        if isinstance(balance_classes, bool):
            self._balance_classes = balance_classes
        else:
            raise ValueError(
                "balance_classes should be boolean, and {} was passed".format(balance_classes))

    @property
    def perturbation(self):
        """ Retrieve the value of the option: Noise percentage

        Returns
        -------
        float

        """
        return self._perturbation

    @perturbation.setter
    def perturbation(self, perturbation):
        """ Set the value of the option: Perturbation.

        Parameters
        ----------
        perturbation: float
            from 0.0 to 1.0.

        """
        if (0.0 <= perturbation) and (perturbation <= 1.0):
            self._perturbation = perturbation
        else:
            raise ValueError(
                "noise percentage should be in [0.0..1.0], and {} was passed".format(perturbation))

    def _prepare_for_use(self):
        self._random_state = check_random_state(self.random_state)
        self._next_class_should_be_zero = False

    def next_sample(self, batch_size=1):
        """ Returns next sample from the stream.

        The sample generation works as follows: The 9 features are generated
        with the random generator, initialized with the seed passed by the
        user. Then, the classification function decides, as a function of all
        the attributes, whether to classify the instance as class 0 or class
        1. The next step is to verify if the classes should be balanced, and
        if so, balance the classes. The last step is to add noise, if the noise
        percentage is higher than 0.0.

        The generated sample will have 9 features and 1 label (it has one
        classification task).

        Parameters
        ----------
        batch_size: int (optional, default=1)
            The number of samples to return.

        Returns
        -------
        tuple or tuple list
            Return a tuple with the features matrix and the labels matrix for
            the batch_size samples that were requested.

        """
        data = np.zeros([batch_size, self.n_features + 1])

        for j in range(batch_size):
            self.sample_idx += 1
            group = 0
            desired_class_found = False
            while not desired_class_found:
                salary = 20000 + 130000 * self._random_state.rand()
                commission = 0 if (salary >= 75000) else (
                    10000 + 75000 * self._random_state.rand())
                age = 20 + self._random_state.randint(61)
                elevel = self._random_state.randint(5)
                car = self._random_state.randint(20)
                zipcode = self._random_state.randint(9)
                hvalue = (9 - zipcode) * 100000 * (0.5 + self._random_state.rand())
                hyears = 1 + self._random_state.randint(30)
                loan = self._random_state.rand() * 500000
                group = self._classification_functions[self.classification_function](salary,
                                                                                     commission,
                                                                                     age, elevel,
                                                                                     car,
                                                                                     zipcode,
                                                                                     hvalue,
                                                                                     hyears, loan)
                if not self.balance_classes:
                    desired_class_found = True
                else:
                    if (self._next_class_should_be_zero and (group == 0)) or \
                            ((not self._next_class_should_be_zero) and (group == 1)):
                        desired_class_found = True
                        self._next_class_should_be_zero = not self._next_class_should_be_zero

            if self.perturbation > 0.0:
                salary = self._perturb_value(salary, 20000, 150000)
                if commission > 0:
                    commission = self._perturb_value(commission, 10000, 75000)
                age = np.round(self._perturb_value(age, 20, 80))
                hvalue = self._perturb_value(hvalue, (9 - zipcode) * 100000, 0, 135000)
                hyears = np.round(self._perturb_value(hyears, 1, 30))
                loan = self._perturb_value(loan, 0, 500000)

            for i in range(9):
                data[j, i] = eval(self.feature_names[i])
            data[j, 9] = group

        self.current_sample_x = data[:, :self.n_features]
        self.current_sample_y = data[:, self.n_features:].flatten().astype(int)

        return self.current_sample_x, self.current_sample_y

    def _perturb_value(self, val, val_min, val_max, val_range=None):
        """
        Perturbs the values of the features by adding noise after assigning a label,
        if the perturbation is higher than 0.0.

        Parameters
        ----------
        val: float
            The value to add noise to.
        val_min: float
            The minimum value after perturbation.
        val_max: float
            The maximum value after perturbation.
        val_range:
            The range of the perturbation.

        Returns
        -------
        float
            The value after perturbation.
        """
        if val_range is None:
            val_range = val_max - val_min
        val += val_range * (2 * (self._random_state.rand() - 0.5)) * self.perturbation
        if val < val_min:
            val = val_min
        elif val > val_max:
            val = val_max
        return val

    def generate_drift(self):
        """
        Generate drift by switching the classification function randomly.

        """
        new_function = self._random_state.randint(10)
        while new_function == self.classification_function:
            new_function = self._random_state.randint(10)
        self.classification_function = new_function

    @staticmethod
    def _classification_function_zero(salary, commission, age, elevel, car, zipcode, hvalue,
                                      hyears, loan):
        """ classification_function_zero

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        return 0 if ((age < 40) or (60 <= age)) else 1

    @staticmethod
    def _classification_function_one(salary, commission, age, elevel, car, zipcode, hvalue, hyears,
                                     loan):
        """ classification_function_one

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        if age < 40:
            return 0 if ((50000 <= salary) and (salary <= 100000)) else 1
        elif age < 60:
            return 0 if ((75000 <= salary) and (salary <= 125000)) else 1
        else:
            return 0 if ((25000 <= salary) and (salary <= 75000)) else 1

    @staticmethod
    def _classification_function_two(salary, commission, age, elevel, car, zipcode, hvalue, hyears,
                                     loan):
        """ classification_function_two

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        if age < 40:
            return 0 if ((elevel == 0) or (elevel == 1)) else 1
        elif age < 60:
            return 0 if ((elevel == 1) or (elevel == 2) or (elevel == 3)) else 1
        else:
            return 0 if ((elevel == 2) or (elevel == 3)) or (elevel == 4) else 1

    @staticmethod
    def _classification_function_three(salary, commission, age, elevel, car, zipcode, hvalue,
                                       hyears, loan):
        """ classification_function_three

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        if age < 40:
            if (elevel == 0) or (elevel == 1):
                return 0 if ((25000 <= salary) and (salary <= 75000)) else 1
            else:
                return 0 if ((50000 <= salary) and (salary <= 100000)) else 1
        elif age < 60:
            if (elevel == 1) or (elevel == 2) or (elevel == 3):
                return 0 if ((50000 <= salary) and (salary <= 100000)) else 1
            else:
                return 0 if ((75000 <= salary) and (salary <= 125000)) else 1
        else:
            if (elevel == 2) or (elevel == 3) or (elevel == 4):
                return 0 if ((50000 <= salary) and (salary <= 100000)) else 1
            else:
                return 0 if ((25000 <= salary) and (salary <= 75000)) else 1

    @staticmethod
    def _classification_function_four(salary, commission, age, elevel, car, zipcode, hvalue,
                                      hyears, loan):
        """ classification_function_four

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        if age < 40:
            if (50000 <= salary) and (salary <= 100000):
                return 0 if ((100000 <= loan) and (loan <= 300000)) else 1
            else:
                return 0 if ((200000 <= salary) and (salary <= 400000)) else 1
        elif age < 60:
            if (75000 <= salary) and (salary <= 125000):
                return 0 if ((200000 <= salary) and (loan <= 400000)) else 1
            else:
                return 0 if ((300000 <= salary) and (salary <= 500000)) else 1
        else:
            if (25000 <= salary) and (salary <= 75000):
                return 0 if ((300000 <= loan) and (loan <= 500000)) else 1
            else:
                return 0 if ((75000 <= loan) and (loan <= 300000)) else 1

    @staticmethod
    def _classification_function_five(salary, commission, age, elevel, car, zipcode, hvalue,
                                      hyears, loan):
        """ classification_function_five

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """

        totalsalary = salary + commission

        if age < 40:
            return 0 if ((50000 <= totalsalary) and (totalsalary <= 100000)) else 1
        elif age < 60:
            return 0 if ((75000 <= totalsalary) and (totalsalary <= 125000)) else 1
        else:
            return 0 if ((25000 <= totalsalary) and (totalsalary <= 75000)) else 1

    @staticmethod
    def _classification_function_six(salary, commission, age, elevel, car, zipcode, hvalue, hyears,
                                     loan):
        """ classification_function_six

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        disposable = (2 * (salary + commission) / 3 - loan / 5 - 20000)
        return 0 if disposable > 1 else 1

    @staticmethod
    def _classification_function_seven(salary, commission, age, elevel, car, zipcode, hvalue,
                                       hyears, loan):
        """ classification_function_seven

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        disposable = (2 * (salary + commission) / 3 - 5000 * elevel - 20000)
        return 0 if disposable > 1 else 1

    @staticmethod
    def _classification_function_eight(salary, commission, age, elevel, car, zipcode, hvalue,
                                       hyears, loan):
        """ classification_function_eight

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        disposable = (2 * (salary + commission) / 3 - 5000 * elevel - loan / 5 - 10000)
        return 0 if disposable > 1 else 1

    @staticmethod
    def _classification_function_nine(salary, commission, age, elevel, car, zipcode, hvalue,
                                      hyears, loan):
        """ classification_function_nine

        Parameters
        ----------
        salary: float
            Numeric feature: Salary.

        commission: float
            Numeric feature: Commission.

        age: int
            Numeric feature: Age.

        elevel: int
            Categorical feature: Education level.

        car: int
            Categorical feature: Car maker.

        zipcode; int
            Categorical feature: Zipcode.

        hvalue: flaot
            Numeric feature: Value of the house.

        hyears: float
            Numeric feature: Years house owned.

        loan: float
            Numeric feature: Total amount of loan.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        equity = 0
        if hyears >= 20:
            equity = hvalue * (hyears - 20) / 10
        disposable = (2 * (salary + commission) / 3 - 5000 * elevel + equity / 5 - 10000)
        return 0 if disposable > 1 else 1



import setuptools, os

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

f = open(os.path.join(os.path.dirname(os.path.realpath(__file__)), 'pandioml/requirements.txt'), "r")
_reqs = f.read().splitlines()

INSTALL_REQUIRES = _reqs
EXTRAS_REQUIRE = {
    "pulsar": ['pulsar-client==2.7.2']
}

setuptools.setup(
    name="pandioml",
    version="1.0.6",
    author="Joshua Odmark",
    author_email="josh@pandio.com",
    description="Pandio's machine learning library.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/pandio-com/pandioml",
    packages=setuptools.find_packages(exclude=['tests']),
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Topic :: Documentation",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    keywords=" ".join(
        [
            "machine learning",
            "data science",
            "pandio",
            "artificial intelligence",
            "build models",
            "online machine learning",
            "streaming",
            "classifiers",
            "supervised learning",
            "regressions",
        ]
    ),
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    setup_requires=['pytest-runner'],
    tests_require=['pytest==4.4.1'],
    test_suite='tests'
)



from river.ensemble.adaptive_random_forest import AdaptiveRandomForestClassifier, AdaptiveRandomForestRegressor
from river.ensemble.bagging import ADWINBaggingClassifier, BaggingClassifier, BaggingRegressor, LeveragingBaggingClassifier
from river.ensemble.boosting import AdaBoostClassifier
from river.ensemble.streaming_random_patches import SRPClassifier


from sys import exc_info
import traceback


class Pipelines:
    _pipelines = {}
    _output = {}
    _error_content = None

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._pipelines = {}
        self._output = {}
        self._error_content = None

    def get_keys(self):
        return list(self._pipelines.keys())

    def builder(self):
        pass

    def add(self, name, pipeline):
        self._pipelines[name] = [pipeline]
        return self

    def go(self, name=None, fnc=None):
        if name is not None:
            try:
                self._output[name] = self.try_catch(self._pipelines[name][0], fnc)
            except Exception as e:
                raise Exception(f"Could not execute pipeline: {e}")
        else:
            for name in self._pipelines:
                try:
                    self._output[name] = self.try_catch(self._pipelines[name][0])
                except Exception as e:
                    raise Exception(f"Could not execute pipeline: {e}")

        return self._output

    def try_catch(self, handler, fnc):
        #try:
        return handler.go(fnc)
        #except Exception as e:
        #    tb = traceback.extract_tb(exc_info()[2])
        #    self._error_content = e, tb
        #    return self._error_content


class Pipeline:
    _steps = []
    _final_step = None
    _done_step = None
    _error_step = None
    _error_content = None
    _fnc = None

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._steps = []
        self._final_step = None
        self._done_step = None
        self._error_step = None
        self._error_content = None

    def go(self, fnc=None):
        self._fnc = fnc
        result = None
        for step in self._steps:
            if self._error_content is not None:
                break
            if result is not None:
                step[2]['result'] = result
            result = self.try_catch(step[0], *step[1], **step[2])
            if self._fnc is not None:
                self._fnc.set_result(result)

        if self._error_content is None:
            if self._final_step is not None and callable(self._final_step[0]):
                if result is not None:
                    self._final_step[2]['result'] = result
                result = self.try_catch(self._final_step[0], *self._final_step[1], **self._final_step[2])
                if self._fnc is not None:
                    self._fnc.set_result(result)

            if self._done_step is not None and callable(self._done_step[0]):
                if result is not None:
                    self._done_step[2]['result'] = result
                return self.try_catch(self._done_step[0], *self._done_step[1], **self._done_step[2])

        if self._error_content is not None:
            self._error_step[2]['result'] = self._error_content
            return self.try_catch(self._error_step[0], *self._error_step[1], **self._error_step[2])

        return None

    def final(self, fnc, *args, **kwargs):
        self._final_step = [fnc, args, kwargs]
        return self

    def done(self, fnc, *args, **kwargs):
        self._done_step = [fnc, args, kwargs]
        return self

    def catch(self, fnc, *args, **kwargs):
        self._error_step = [fnc, args, kwargs]
        return self

    def then(self, fnc, *args, **kwargs):
        self._steps.append([fnc, args, kwargs])
        return self

    def try_catch(self, handler, *args, **kwargs):
        try:
            return handler(*args, **kwargs)
        except Exception as e:
            tb = traceback.extract_tb(exc_info()[2])
            self._error_content = e, tb
            return self._error_content



from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, Float, Integer
from .scikit_multiflow_classes import SKStream, check_random_state
import numpy as np


class SineGeneratorDataset(Stream):
    """
    A generator for data regarding sine values with the ability to balance and add noise.

    Each record is an instance of Sine

    class Sine(Record):
        sine = Float()
        sine_reverse = Float()
        sine_two = Float()
        sine_two_reverse = Float()

    Sine stream generator.

    This generator is an implementation of the dara stream with abrupt
    concept drift, as described in Gama, Joao, et al [1]_.
    It generates up to 4 relevant numerical attributes, that vary from 0 to 1,
    where only 2 of them are relevant to the classification task and the other
    2 are added by request of the user. A classification function is chosen
    among four possible ones:
    0. SINE1. Abrupt concept drift, noise-free examples. It has two relevant
       attributes. Each attributes has values uniformly distributed in [0; 1].
       In the first context all points below the curve :math:`y = sin(x)` are
       classified as positive.
    1. Reversed SINE1. The reversed classification of SINE1.
    2. SINE2. The same two relevant attributes. The classification function
       is :math:`y < 0.5 + 0.3 sin(3 \pi  x)`.
    3. Reversed SINE2. The reversed classification of SINE2.
    Concept drift can be introduced by changing the classification function.
    This can be done manually or using ``ConceptDriftStream``.
    Two important features are the possibility to balance classes, which
    means the class distribution will tend to a uniform one, and the possibility
    to add noise, which will, add two non relevant attributes.
    Parameters
    ----------
    classification_function: int (Default: 0)
        Which of the four classification functions to use for the generation.
        From 0 to 3.
    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.
    balance_classes: bool (Default: False)
        Whether to balance classes or not. If balanced, the class
        distribution will converge to a uniform distribution.
    has_noise: bool (Default: False)
        Adds 2 non relevant features to the stream.
    References
    ----------
    .. [1] Gama, Joao, et al.'s 'Learning with drift
       detection.' Advances in artificial intelligence–SBIA 2004. Springer Berlin
       Heidelberg, 2004. 286-295."

    """
    dataset = None

    def __init__(self, *args, **kwargs):
        self.dataset = SineGenerator(*args, **kwargs)

    def next(self):
        features, label = self.dataset.next_sample()

        dict = {'label': label[0].item()}

        keys = list(Sine._fields.keys())

        index = 0
        for element in features[0]:
            dict[keys[index]] = element.item()
            index += 1

        return Sine(**dict)

    @staticmethod
    def schema():
        return JsonSchema(Sine)


class Sine(Record):
    sine = Float()
    sine_reverse = Float()
    label = Integer()


class SineGenerator(SKStream):
    r""" Sine stream generator.

    This generator is an implementation of the dara stream with abrupt
    concept drift, as described in Gama, Joao, et al [1]_.

    It generates up to 4 relevant numerical attributes, that vary from 0 to 1,
    where only 2 of them are relevant to the classification task and the other
    2 are added by request of the user. A classification function is chosen
    among four possible ones:

    0. SINE1. Abrupt concept drift, noise-free examples. It has two relevant
       attributes. Each attributes has values uniformly distributed in [0; 1].
       In the first context all points below the curve :math:`y = sin(x)` are
       classified as positive.
    1. Reversed SINE1. The reversed classification of SINE1.
    2. SINE2. The same two relevant attributes. The classification function
       is :math:`y < 0.5 + 0.3 sin(3 \pi  x)`.
    3. Reversed SINE2. The reversed classification of SINE2.

    Concept drift can be introduced by changing the classification function.
    This can be done manually or using ``ConceptDriftStream``.

    Two important features are the possibility to balance classes, which
    means the class distribution will tend to a uniform one, and the possibility
    to add noise, which will, add two non relevant attributes.

    Parameters
    ----------
    classification_function: int (Default: 0)
        Which of the four classification functions to use for the generation.
        From 0 to 3.

    random_state: int, RandomState instance or None, optional (default=None)
        If int, random_state is the seed used by the random number generator;
        If RandomState instance, random_state is the random number generator;
        If None, the random number generator is the RandomState instance used
        by `np.random`.

    balance_classes: bool (Default: False)
        Whether to balance classes or not. If balanced, the class
        distribution will converge to a uniform distribution.

    has_noise: bool (Default: False)
        Adds 2 non relevant features to the stream.

    References
    ----------
    .. [1] Gama, Joao, et al.'s 'Learning with drift
       detection.' Advances in artificial intelligence–SBIA 2004. Springer Berlin
       Heidelberg, 2004. 286-295."

    Examples
    --------
    >>> # Imports
    >>> from skmultiflow.data.sine_generator import SineGenerator
    >>> # Setting up the stream
    >>> stream = SineGenerator(classification_function = 2, random_state = 112,
    ...  balance_classes = False, has_noise = True)
    >>> # Retrieving one sample
    >>> stream.next_sample()
    (array([[0.37505713, 0.64030462, 0.95001658, 0.0756772 ]]), array([1.]))
    >>> stream.next_sample(10)
    (array([[0.77692966, 0.83274576, 0.05480574, 0.81767738],
       [0.88535146, 0.72234651, 0.00255603, 0.98119928],
       [0.34341985, 0.09475989, 0.39464259, 0.00494492],
       [0.73670683, 0.95580687, 0.82060937, 0.344983  ],
       [0.37854446, 0.78476361, 0.08623151, 0.54607394],
       [0.16222602, 0.29006973, 0.04500817, 0.33218776],
       [0.73653322, 0.83921149, 0.70936161, 0.18840112],
       [0.98566856, 0.38800331, 0.50315448, 0.76353033],
       [0.68373245, 0.72195738, 0.21415209, 0.76309258],
       [0.07521616, 0.6108907 , 0.42563042, 0.23435109]]),
       array([1., 0., 1., 0., 1., 1., 1., 0., 0., 1.]))
    >>> stream.n_remaining_samples()
    -1
    >>> stream.has_more_samples()
    True

    """
    _NUM_BASE_ATTRIBUTES = 2
    _TOTAL_ATTRIBUTES_INCLUDING_NOISE = 4

    def __init__(self, classification_function=0, random_state=None, balance_classes=False,
                 has_noise=False):
        super().__init__()

        # Classification functions to use
        self._classification_functions = [self._classification_function_zero,
                                          self._classification_function_one,
                                          self._classification_function_two,
                                          self._classification_function_three]
        self.classification_function = classification_function
        self.random_state = random_state
        self.has_noise = has_noise
        self.balance_classes = balance_classes
        self.n_num_features = self._NUM_BASE_ATTRIBUTES
        self.n_classes = 2
        self.n_targets = 1
        self._random_state = None  # This is the actual random_state object used internally
        self.next_class_should_be_zero = False
        self.name = "Sine Generator"

        if self.has_noise:
            self.n_num_features = self._TOTAL_ATTRIBUTES_INCLUDING_NOISE
        self.n_features = self.n_num_features
        self.target_names = ["target_0"]
        self.feature_names = ["att_num_" + str(i) for i in range(self.n_features)]
        self.target_values = [i for i in range(self.n_classes)]

        self._prepare_for_use()

    @property
    def classification_function(self):
        """ Retrieve the index of the current classification function.

        Returns
        -------
        int
            index of the classification function [0,1,2,3]
        """
        return self._classification_function_idx

    @classification_function.setter
    def classification_function(self, classification_function_idx):
        """ Set the index of the current classification function.

        Parameters
        ----------
        classification_function_idx: int (0,1,2,3)
        """
        if classification_function_idx in range(4):
            self._classification_function_idx = classification_function_idx
        else:
            raise ValueError("classification_function takes only these "
                             "values: 0, 1, 2, 3, and {} was "
                             "passed".format(classification_function_idx))

    @property
    def balance_classes(self):
        """ Retrieve the value of the option: Balance classes

        Returns
        -------
        Boolean
            True is the classes are balanced
        """
        return self._balance_classes

    @balance_classes.setter
    def balance_classes(self, balance_classes):
        """ Set the value of the option: Balance classes.

        Parameters
        ----------
        balance_classes: Boolean

        """
        if isinstance(balance_classes, bool):
            self._balance_classes = balance_classes
        else:
            raise ValueError("balance_classes should be boolean,"
                             " and {} was passed".format(balance_classes))

    @property
    def has_noise(self):
        """ Retrieve the value of the option: add noise.

        Returns
        -------
        Boolean
            True is the noise is added
        """
        return self._has_noise

    @has_noise.setter
    def has_noise(self, has_noise):
        """ Set the value of the option: add noise.

        Parameters
        ----------
        has_noise: Boolean

        """
        if isinstance(has_noise, bool):
            self._has_noise = has_noise
        else:
            raise ValueError("has_noise should be boolean, {} was passed".format(has_noise))

    def _prepare_for_use(self):
        self._random_state = check_random_state(self.random_state)
        self.next_class_should_be_zero = False

    def next_sample(self, batch_size=1):
        """ Returns next sample from the stream.

        The sample generation works as follows: The two attributes are
        generated with the random generator, initialized with the seed passed
        by the user. Then, the classification function decides whether to
        classify the instance as class 0 or class 1. The next step is to
        verify if the classes should be balanced, and if so, balance the
        classes. The last step is to add noise, if the has_noise is True.

        The generated sample will have 2 relevant features, and an additional
        two noise features if option chosen, and 1 label (it has one
        classification task).

        Parameters
        ----------
        batch_size: int (optional, default=1)
            The number of samples to return.

        Returns
        -------
        tuple or tuple list
            Return a tuple with the features matrix and the labels matrix for
            the batch_size samples that were requested.

        """

        data = np.zeros([batch_size, self.n_features + 1])

        for j in range(batch_size):
            self.sample_idx += 1
            att1 = att2 = 0.0
            group = 0
            desired_class_found = False
            while not desired_class_found:
                att1 = self._random_state.rand()
                att2 = self._random_state.rand()
                group = self._classification_functions[self.classification_function](att1, att2)

                if not self.balance_classes:
                    desired_class_found = True
                else:
                    if (self.next_class_should_be_zero and (group == 0)) or \
                            ((not self.next_class_should_be_zero) and (group == 1)):
                        desired_class_found = True
                        self.next_class_should_be_zero = not self.next_class_should_be_zero

            data[j, 0] = att1
            data[j, 1] = att2

            if self.has_noise:
                for i in range(self._NUM_BASE_ATTRIBUTES, self._TOTAL_ATTRIBUTES_INCLUDING_NOISE):
                    data[j, i] = self._random_state.rand()
                data[j, 4] = group
            else:
                data[j, 2] = group

        self.current_sample_x = data[:, :self.n_features]
        self.current_sample_y = data[:, self.n_features:].flatten().astype(int)

        return self.current_sample_x, self.current_sample_y

    def generate_drift(self):
        """
        Generate drift by switching the classification function randomly.

        """
        new_function = self._random_state.randint(4)
        while new_function == self.classification_function:
            new_function = self._random_state.randint(4)
        self.classification_function = new_function

    @staticmethod
    def _classification_function_zero(att1, att2):
        """ classification_function_zero

        Decides the sample class label based on SINE1 function.


        Parameters
        ----------
        att1: float
            First numeric attribute.

        att2: float
            Second numeric attribute.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        return 0 if (att1 >= np.sin(att2)) else 1

    @staticmethod
    def _classification_function_one(att1, att2):
        """ classification_function_one

        Decides the sample class label based on reversed SINE1 function.

        Parameters
        ----------
        att1: float
            First numeric attribute.

        att2: float
            Second numeric attribute.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        return 0 if (att1 < np.sin(att2)) else 1

    @staticmethod
    def _classification_function_two(att1, att2):
        """ classification_function_two

        Decides the sample class label based on SINE2 function.

        Parameters
        ----------
        att1: float
            First numeric attribute.

        att2: float
            Second numeric attribute.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        return 0 if (att1 >= 0.5 + 0.3 * np.sin(3 * np.pi * att2)) else 1

    @staticmethod
    def _classification_function_three(att1, att2):
        """ classification_function_three

        Decides the sample class label based on reversed SINE2 function.

        Parameters
        ----------
        att1: float
            First numeric attribute.

        att2: float
            Second numeric attribute.

        Returns
        -------
        int
            Returns the sample class label, either 0 or 1.

        """
        return 0 if (att1 < 0.5 + 0.3 * np.sin(3 * np.pi * att2)) else 1



from pandioml.data.stream import Stream
import pulsar
from pandioml.data.record import JsonSchema, Record, Float, Integer


class WebHostingDataset(Stream):
    """
    Contains 12,496,728 server resource metric events recorded over a 3 month period of time.

    Each record is an instance of ResourceEvent

    class ResourceEvent(Record):
        cpu_usage_mhz = Float()
        cpu_usage_percent = Float()
        cpu_provisioned_mhz = Float()
        ingress_bandwidth_kbs = Float()
        egress_bandwidth_kbs = Float()
        memory_kb = Float()
        timestamp_ms = Integer()
        weekday = Integer()
        weekend = Integer()
        month = Integer()
        day = Integer()
        hour = Integer()
        cpu_prev_usage_percent = Float()
        cpu_diff_usage_percent = Float()
        ingress_bandwidth_prev_kbs = Float()
        ingress_bandwidth_diff_kbs = Float()
        egress_bandwidth_prev_kbs = Float()
        egress_bandwidth_diff_kbs = Float()
    """

    index = 0
    dataset = None

    def __init__(self, start_id=pulsar.MessageId.earliest, receiver_queue_size=1000, pandio_token=None):
        self.client = pulsar.Client('pulsar+ssl://joshuaeric--gray-guan.us-west2.gcp.pulsar.pandio.com:6651',
                       authentication=pulsar.AuthenticationToken(pandio_token))

        self.reader = self.client.create_reader('persistent://public/default/shared-webhost', start_id,
                                                schema=JsonSchema(ResourceEvent),
                                                receiver_queue_size=receiver_queue_size)

    def next(self):
        msg = self.reader.read_next()
        self.index += 1
        return msg

    @staticmethod
    def schema():
        return JsonSchema(ResourceEvent)


class ResourceEvent(Record):
    cpu_usage_mhz = Float()
    cpu_usage_percent = Float()
    cpu_provisioned_mhz = Float()
    ingress_bandwidth_kbs = Float()
    egress_bandwidth_kbs = Float()
    memory_kb = Float()
    timestamp_ms = Integer()
    weekday = Integer()
    weekend = Integer()
    month = Integer()
    day = Integer()
    hour = Integer()
    cpu_prev_usage_percent = Float()
    cpu_diff_usage_percent = Float()
    ingress_bandwidth_prev_kbs = Float()
    ingress_bandwidth_diff_kbs = Float()
    egress_bandwidth_prev_kbs = Float()
    egress_bandwidth_diff_kbs = Float()



import faker
from pandioml.data.stream import Stream
from pandioml.data.record import JsonSchema, Record, String, Integer


class FormSubmissionGenerator(Stream):
    """
    Web Form submission generator using the Faker class.

    Each record is an instance of Submission

    class Submission(Record):
        email = String()
        ip = String()
        timestamp = Float()
    """
    dataset = None

    def __init__(self):
        self.dataset = faker.Faker('en_US')

    def next(self):
        return Submission(email=getattr(self.dataset, 'ascii_email')(), ip=getattr(self.dataset, 'ipv4')(),
                          timestamp=getattr(self.dataset, 'unix_time')())

    @staticmethod
    def schema():
        return JsonSchema(Submission)


class Submission(Record):
    email = String()
    ip = String()
    timestamp = Integer()



from river.metrics.cluster.base import InternalMetric
from river.metrics.cluster.bic import BIC
from river.metrics.cluster.daviesbouldin import DaviesBouldin
from river.metrics.cluster.generalized_dunn import GD43, GD53
from river.metrics.cluster.i_index import IIndex
from river.metrics.cluster.ps import PS
from river.metrics.cluster.r2 import R2 as CR2
from river.metrics.cluster.rmsstd import MSSTD, RMSSTD
from river.metrics.cluster.sd_validation import SD
from river.metrics.cluster.separation import Separation
from river.metrics.cluster.silhouette import Silhouette
from river.metrics.cluster.ssb import SSB
from river.metrics.cluster.ssq_based import WB, CalinskiHarabasz, Hartigan
from river.metrics.cluster.ssw import SSW, BallHall, Cohesion, Xu
from river.metrics.cluster.xiebeni import XieBeni
from river.metrics.accuracy import Accuracy
from river.metrics.balanced_accuracy import BalancedAccuracy
from river.metrics.base import (
    BinaryMetric,
    ClassificationMetric,
    Metric,
    Metrics,
    MultiClassMetric,
    MultiOutputClassificationMetric,
    MultiOutputRegressionMetric,
    RegressionMetric,
    WrapperMetric
)
from river.metrics.cross_entropy import CrossEntropy
from river.metrics.exact_match import ExactMatch
from river.metrics.fbeta import (
    F1,
    ExampleF1,
    ExampleFBeta,
    FBeta,
    MacroF1,
    MacroFBeta,
    MicroF1,
    MicroFBeta,
    MultiFBeta,
    WeightedF1,
    WeightedFBeta,
)
from river.metrics.geometric_mean import GeometricMean
from river.metrics.hamming import Hamming, HammingLoss
from river.metrics.jaccard import Jaccard
from river.metrics.kappa import CohenKappa, KappaM, KappaT
from river.metrics.log_loss import LogLoss
from river.metrics.mae import MAE
from river.metrics.mcc import MCC
from river.metrics.mse import MSE, RMSE, RMSLE
from river.metrics.multioutput import RegressionMultiOutput
from river.metrics.precision import (
    ExamplePrecision,
    MacroPrecision,
    MicroPrecision,
    Precision,
    WeightedPrecision,
)
from river.metrics.r2 import R2
from river.metrics.recall import ExampleRecall, MacroRecall, MicroRecall, Recall, WeightedRecall
from river.metrics.report import ClassificationReport
from river.metrics.roc_auc import ROCAUC
from river.metrics.rolling import Rolling
from river.metrics.smape import SMAPE
from river.metrics.time_rolling import TimeRolling
from river.metrics.confusion import ConfusionMatrix, MultiLabelConfusionMatrix


__all__ = [
    "Accuracy",
    "BalancedAccuracy",
    "BinaryMetric",
    "ClassificationMetric",
    "ClassificationReport",
    "CohenKappa",
    "CrossEntropy",
    "ExactMatch",
    "ExamplePrecision",
    "ExampleRecall",
    "ExampleF1",
    "ExampleFBeta",
    "F1",
    "FBeta",
    "GeometricMean",
    "Hamming",
    "HammingLoss",
    "Jaccard",
    "KappaM",
    "KappaT",
    "LogLoss",
    "MAE",
    "MacroF1",
    "MacroFBeta",
    "MacroPrecision",
    "MacroRecall",
    "MCC",
    "Metric",
    "Metrics",
    "MicroF1",
    "MicroFBeta",
    "MicroPrecision",
    "MicroRecall",
    "MultiClassMetric",
    "MultiFBeta",
    "MultiOutputClassificationMetric",
    "MultiOutputRegressionMetric",
    "MSE",
    "Precision",
    "Recall",
    "RegressionMetric",
    "RegressionMultiOutput",
    "RMSE",
    "RMSLE",
    "ROCAUC",
    "Rolling",
    "R2",
    "SMAPE",
    "TimeRolling",
    "WeightedF1",
    "WeightedFBeta",
    "WeightedPrecision",
    "WeightedRecall",
    "WrapperMetric",
    "BallHall",
    "BIC",
    "CalinskiHarabasz",
    "Cohesion",
    "DaviesBouldin",
    "GD43",
    "GD53",
    "Hartigan",
    "IIndex",
    "InternalMetric",
    "MSSTD",
    "PS",
    "CR2",
    "RMSSTD",
    "SD",
    "Separation",
    "Silhouette",
    "SSB",
    "SSW",
    "XieBeni",
    "WB",
    "Xu",
    "ConfusionMatrix",
    "MultiLabelConfusionMatrix"
]



from river.feature_selection.k_best import SelectKBest
from river.feature_selection.random import PoissonInclusion
from river.feature_selection.variance import VarianceThreshold

__all__ = ["PoissonInclusion", "SelectKBest", "VarianceThreshold"]